import re

TEMP_PATH_TEMPLATE = "gamdl_temp_{}"
ILLEGAL_CHARS_RE = r'[\\/:*?"<>|;]'
ILLEGAL_CHAR_REPLACEMENT = "_"

SONG_MEDIA_TYPE = {"song", "songs", "library-songs"}
ALBUM_MEDIA_TYPE = {"album", "albums", "library-albums"}
MUSIC_VIDEO_MEDIA_TYPE = {"music-video", "music-videos", "library-music-videos"}
ARTIST_MEDIA_TYPE = {"artist", "artists", "library-artists"}
UPLOADED_VIDEO_MEDIA_TYPE = {"post", "uploaded-videos"}
PLAYLIST_MEDIA_TYPE = {"playlist", "playlists", "library-playlists"}

ARTIST_AUTO_SELECT_KEY_MAP = {
    "main-albums": ("views", "full-albums"),
    "compilation-albums": ("views", "compilation-albums"),
    "live-albums": ("views", "live-albums"),
    "singles-eps": ("views", "singles"),
    "all-albums": ("relationships", "albums"),
    "top-songs": ("views", "top-songs"),
    "music-videos": ("relationships", "music-videos"),
}
ARTIST_AUTO_SELECT_STR_MAP = {
    "main-albums": "Main Albums",
    "compilation-albums": "Compilation Albums",
    "live-albums": "Live Albums",
    "singles-eps": "Singles & EPs",
    "all-albums": "All Albums",
    "top-songs": "Top Songs",
    "music-videos": "Music Videos",
}

VALID_URL_PATTERN = re.compile(
    r"https://(?:classical\.)?music\.apple\.com"
    r"(?:"
    r"/(?P<storefront>[a-z]{2})"
    r"/(?P<type>artist|album|playlist|song|music-video|post)"
    r"(?:/(?P<slug>[^\s/]+))?"
    r"/(?P<id>[0-9]+|pl\.[0-9a-z]{32}|pl\.u-[a-zA-Z0-9]+)"
    r"(?:\?i=(?P<sub_id>[0-9]+))?"
    r"|"
    r"(?:/(?P<library_storefront>[a-z]{2}))?"
    r"/library/(?P<library_type>playlist|albums)"
    r"/(?P<library_id>p\.[a-zA-Z0-9]+|l\.[a-zA-Z0-9]+)"
    r")"
)



from enum import Enum

from .constants import (
    ARTIST_AUTO_SELECT_KEY_MAP,
    ARTIST_AUTO_SELECT_STR_MAP,
)


class DownloadMode(Enum):
    YTDLP = "ytdlp"
    NM3U8DLRE = "nm3u8dlre"


class RemuxMode(Enum):
    FFMPEG = "ffmpeg"
    MP4BOX = "mp4box"


class RemuxFormatMusicVideo(Enum):
    M4V = "m4v"
    MP4 = "mp4"


class ArtistAutoSelect(Enum):
    MAIN_ALBUMS = "main-albums"
    COMPILATION_ALBUMS = "compilation-albums"
    LIVE_ALBUMS = "live-albums"
    SINGLES_EPS = "singles-eps"
    ALL_ALBUMS = "all-albums"
    TOP_SONGS = "top-songs"
    MUSIC_VIDEOS = "music-videos"

    @property
    def path_key(self) -> tuple[str, str]:
        return ARTIST_AUTO_SELECT_KEY_MAP[self.value]

    def __str__(self) -> str:
        return ARTIST_AUTO_SELECT_STR_MAP[self.value]



import logging
import urllib.parse

import m3u8
from async_lru import alru_cache
from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from pywidevine import Cdm

from ..utils import get_response
from .constants import MP4_FORMAT_CODECS
from .enums import MediaRating, MediaType, MusicVideoCodec, MusicVideoResolution
from .interface import AppleMusicInterface
from .types import DecryptionKeyAv, MediaFileFormat, MediaTags, StreamInfo, StreamInfoAv

logger = logging.getLogger(__name__)


class AppleMusicMusicVideoInterface(AppleMusicInterface):
    def __init__(self, interface: AppleMusicInterface):
        self.__dict__.update(interface.__dict__)

    async def get_itunes_page_metadata(
        self,
        music_video_metadata: dict,
    ) -> dict:
        alt_id = self.get_alt_id(music_video_metadata)
        itunes_page = await self.itunes_api.get_itunes_page(
            "music-video",
            alt_id,
        )
        return itunes_page["storePlatformData"]["product-dv"]["results"][alt_id]

    def get_m3u8_master_url_from_webplayback(self, webplayback: dict) -> str:
        m3u8_master_url = webplayback["hls-playlist-url"]
        return m3u8_master_url

    def get_m3u8_master_url_from_itunes_page_metadata(
        self,
        itunes_page_metadata: dict,
    ) -> dict:
        stream_url = itunes_page_metadata["offers"][0]["assets"][0]["hlsUrl"]

        url_parts = urllib.parse.urlparse(stream_url)
        query = urllib.parse.parse_qs(url_parts.query, keep_blank_values=True)
        query.update({"aec": "HD", "dsid": "1"})

        m3u8_master_url = url_parts._replace(
            query=urllib.parse.urlencode(query, doseq=True)
        ).geturl()

        return m3u8_master_url

    def get_alt_id(self, metadata: dict) -> str | None:
        music_video_url = metadata["attributes"].get("url")
        if music_video_url is None:
            return None

        alt_id = music_video_url.split("/")[-1].split("?")[0]
        logger.debug(f"Alt ID: {alt_id}")

        return alt_id

    @alru_cache()
    async def get_album(
        self,
        collection_id: int,
    ) -> dict | None:
        album_response = await self.apple_music_api.get_album(collection_id)
        if not album_response:
            return None
        return album_response["data"][0]

    async def get_tags(
        self,
        metadata: dict,
        itunes_page_metadata: dict,
    ) -> MediaTags:
        alt_id = self.get_alt_id(metadata)
        lookup_metadata = (await self.itunes_api.get_lookup_result(alt_id))["results"]

        explicitness = lookup_metadata[0]["trackExplicitness"]
        if explicitness == "notExplicit":
            rating = MediaRating.NONE
        elif explicitness == "explicit":
            rating = MediaRating.EXPLICIT
        else:
            rating = MediaRating.CLEAN

        tags = MediaTags(
            artist=lookup_metadata[0]["artistName"],
            artist_id=int(lookup_metadata[0]["artistId"]),
            copyright=itunes_page_metadata.get("copyright"),
            date=self.parse_date(lookup_metadata[0]["releaseDate"]),
            genre=lookup_metadata[0]["primaryGenreName"],
            genre_id=int(itunes_page_metadata["genres"][0]["genreId"]),
            media_type=MediaType.MUSIC_VIDEO,
            storefront=int(self.itunes_api.storefront_id.split("-")[0]),
            title=lookup_metadata[0]["trackCensoredName"],
            title_id=int(metadata["id"]),
            rating=rating,
        )

        if len(lookup_metadata) > 1:
            album = await self.get_album(itunes_page_metadata["collectionId"])
            if not album:
                return tags

            tags.album = lookup_metadata[1]["collectionCensoredName"]
            tags.album_artist = lookup_metadata[1]["artistName"]
            tags.album_id = int(itunes_page_metadata["collectionId"])
            tags.disc = lookup_metadata[0]["discNumber"]
            tags.disc_total = lookup_metadata[0]["discCount"]
            tags.compilation = album["attributes"]["isCompilation"]
            tags.track = lookup_metadata[0]["trackNumber"]
            tags.track_total = lookup_metadata[0]["trackCount"]

        logger.debug(f"Tags: {tags}")

        return tags

    async def get_stream_info(
        self,
        metadata: dict,
        itunes_page_metadata: dict,
        codec_priority: list[MusicVideoCodec],
        resolution: MusicVideoResolution,
    ) -> StreamInfoAv:
        alt_video_id = self.get_alt_id(metadata)
        if alt_video_id == metadata["id"]:
            m3u8_master_url = self.get_m3u8_master_url_from_itunes_page_metadata(
                itunes_page_metadata,
            )
        else:
            webplayback_response = await self.apple_music_api.get_webplayback(
                metadata["id"]
            )
            m3u8_master_url = self.get_m3u8_master_url_from_webplayback(
                webplayback_response["songList"][0],
            )

        playlist_master_m3u8_obj = m3u8.loads(
            (await get_response(m3u8_master_url)).text
        )
        playlist_master_m3u8_obj.base_uri = m3u8_master_url.rpartition("/")[0]
        stream_info_video = await self.get_stream_info_video(
            playlist_master_m3u8_obj,
            codec_priority,
            resolution,
        )
        stream_info_audio = await self.get_stream_info_audio(
            playlist_master_m3u8_obj.data,
            codec_priority,
        )
        if not stream_info_video or not stream_info_audio:
            return None

        use_mp4 = any(
            stream_info_video.codec.startswith(codec) for codec in MP4_FORMAT_CODECS
        ) or any(
            stream_info_audio.codec.startswith(codec) for codec in MP4_FORMAT_CODECS
        )
        if use_mp4:
            file_format = MediaFileFormat.MP4
        else:
            file_format = MediaFileFormat.M4V

        stream_info = StreamInfoAv(
            video_track=stream_info_video,
            audio_track=stream_info_audio,
            file_format=file_format,
        )
        logger.debug(f"Stream info: {stream_info}")

        return stream_info

    def get_video_playlist_from_resolution(
        self,
        video_playlists: list[m3u8.Playlist],
        codec_priority: list[MusicVideoCodec],
        resolution: MusicVideoResolution,
    ) -> m3u8.Playlist | None:
        playlist_results = []
        for codec_index, codec in enumerate(codec_priority):
            for playlist in video_playlists:
                if playlist.stream_info.codecs.startswith(codec.fourcc()):
                    playlist_results.append((codec_index, playlist))

        if not playlist_results:
            return None

        def sort_key(
            item: tuple[int, m3u8.Playlist],
        ) -> tuple[bool, int, int, int, int]:
            codec_index, playlist = item
            playlist_resolution = playlist.stream_info.resolution[-1]
            bandwidth = playlist.stream_info.bandwidth
            exceeds_resolution = playlist_resolution > int(resolution)
            resolution_difference = abs(playlist_resolution - int(resolution))

            return (
                exceeds_resolution,
                resolution_difference,
                codec_index,
                -playlist_resolution,
                -bandwidth,
            )

        playlist_results.sort(key=sort_key)
        return playlist_results[0][1]

    def get_best_stereo_audio_playlist(
        self,
        playlist_master_data: dict,
    ) -> dict | None:
        audio_playlist = next(
            (
                media
                for media in playlist_master_data["media"]
                if media["group_id"] == "audio-stereo-256"
            ),
            None,
        )
        return audio_playlist

    async def get_video_playlist_from_user(
        self,
        video_playlists: list[m3u8.Playlist],
    ) -> m3u8.Playlist:
        choices = [
            Choice(
                name=" | ".join(
                    [
                        playlist.stream_info.codecs[:4],
                        "x".join(str(v) for v in playlist.stream_info.resolution),
                        str(playlist.stream_info.bandwidth),
                    ]
                ),
                value=playlist,
            )
            for playlist in video_playlists
        ]
        selected = await inquirer.select(
            message="Select which video codec to download: (Codec | Resolution | Bitrate)",
            choices=choices,
        ).execute_async()

        return selected

    async def get_audio_playlist_from_user(
        self,
        playlist_master_data: dict,
    ) -> dict:
        choices = [
            Choice(
                name=playlist["group_id"],
                value=playlist,
            )
            for playlist in playlist_master_data["media"]
            if playlist.get("uri")
        ]
        selected = await inquirer.select(
            message="Select which audio codec to download:",
            choices=choices,
        ).execute_async()

        return selected

    def _get_key_by_format(
        self,
        m3u8_obj: m3u8.M3U8,
        key_format: str,
    ) -> str:
        return next(
            (key for key in m3u8_obj.keys if key.keyformat == key_format),
            None,
        ).uri

    def get_widevine_pssh(self, m3u8_obj: m3u8.M3U8) -> str:
        return self._get_key_by_format(
            m3u8_obj,
            "urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed",
        )

    def get_playready_pssh(self, m3u8_obj: m3u8.M3U8) -> str:
        return self._get_key_by_format(
            m3u8_obj,
            "com.microsoft.playready",
        )

    def get_fairplay_key(self, m3u8_obj: m3u8.M3U8) -> str:
        return self._get_key_by_format(
            m3u8_obj,
            "com.apple.streamingkeydelivery",
        )

    async def get_stream_info_video(
        self,
        playlist_master_m3u8_obj: m3u8.M3U8,
        codec_priority: list[MusicVideoCodec],
        resolution: MusicVideoResolution,
    ) -> StreamInfo | None:
        stream_info = StreamInfo()

        if MusicVideoCodec.ASK not in codec_priority:
            playlist = self.get_video_playlist_from_resolution(
                playlist_master_m3u8_obj.playlists,
                codec_priority,
                resolution,
            )
        else:
            playlist = await self.get_video_playlist_from_user(
                playlist_master_m3u8_obj.playlists
            )

        if not playlist:
            return None

        stream_info.stream_url = playlist.uri
        stream_info.codec = playlist.stream_info.codecs
        stream_info.width, stream_info.height = playlist.stream_info.resolution

        playlist_m3u8_obj = m3u8.loads(
            (await get_response(stream_info.stream_url)).text
        )
        stream_info.widevine_pssh = self.get_widevine_pssh(playlist_m3u8_obj)
        stream_info.fairplay_key = self.get_fairplay_key(playlist_m3u8_obj)
        stream_info.playready_pssh = self.get_playready_pssh(playlist_m3u8_obj)

        return stream_info

    async def get_stream_info_audio(
        self,
        playlist_master_data: dict,
        codec_priority: list[MusicVideoCodec],
    ) -> StreamInfo | None:
        stream_info = StreamInfo()

        if MusicVideoCodec.ASK not in codec_priority:
            playlist = self.get_best_stereo_audio_playlist(playlist_master_data)
        else:
            playlist = await self.get_audio_playlist_from_user(playlist_master_data)

        if not playlist:
            return None

        stream_info.stream_url = playlist["uri"]
        stream_info.codec = playlist["group_id"]

        playlist_m3u8_obj = m3u8.loads(
            (await get_response(stream_info.stream_url)).text
        )
        stream_info.widevine_pssh = self.get_widevine_pssh(playlist_m3u8_obj)
        stream_info.fairplay_key = self.get_fairplay_key(playlist_m3u8_obj)
        stream_info.playready_pssh = self.get_playready_pssh(playlist_m3u8_obj)

        return stream_info

    async def get_decryption_key(
        self,
        stream_info: StreamInfoAv,
        cdm: Cdm,
    ) -> DecryptionKeyAv:
        decryption_key_video = await AppleMusicInterface.get_decryption_key(
            self,
            stream_info.video_track.widevine_pssh,
            stream_info.media_id,
            cdm,
        )
        decryption_key_audio = await AppleMusicInterface.get_decryption_key(
            self,
            stream_info.audio_track.widevine_pssh,
            stream_info.media_id,
            cdm,
        )

        return DecryptionKeyAv(
            video_track=decryption_key_video,
            audio_track=decryption_key_audio,
        )



from pathlib import Path

from ..interface.enums import CoverFormat, UploadedVideoQuality
from ..interface.interface_uploaded_video import AppleMusicUploadedVideoInterface
from .downloader_base import AppleMusicBaseDownloader
from .types import DownloadItem


class AppleMusicUploadedVideoDownloader(AppleMusicBaseDownloader):
    def __init__(
        self,
        base_downloader: AppleMusicBaseDownloader,
        interface: AppleMusicUploadedVideoInterface,
        quality: UploadedVideoQuality = UploadedVideoQuality.BEST,
    ):
        self.__dict__.update(base_downloader.__dict__)
        self.interface = interface
        self.quality = quality

    def get_cover_path(self, final_path: str, file_extension: str) -> str:
        return str(Path(final_path).with_suffix(file_extension))

    async def get_download_item(
        self,
        uploaded_video_metadata: dict,
    ) -> DownloadItem:
        try:
            return await self._get_download_item(
                uploaded_video_metadata,
            )
        except Exception as e:
            return DownloadItem(
                media_metadata=uploaded_video_metadata,
                error=e,
            )

    async def _get_download_item(
        self,
        uploaded_video_metadata: dict,
    ) -> DownloadItem:
        download_item = DownloadItem()

        download_item.media_metadata = uploaded_video_metadata

        download_item.media_tags = self.interface.get_tags(
            uploaded_video_metadata,
        )

        download_item.stream_info = await self.interface.get_stream_info(
            uploaded_video_metadata,
            self.quality,
        )

        download_item.random_uuid = self.get_random_uuid()
        download_item.staged_path = self.get_temp_path(
            uploaded_video_metadata["id"],
            download_item.random_uuid,
            "staged",
            "." + download_item.stream_info.file_format.value,
        )
        download_item.final_path = self.get_final_path(
            download_item.media_tags,
            Path(download_item.staged_path).suffix,
            None,
        )

        download_item.cover_url_template = self.interface.get_cover_url_template(
            uploaded_video_metadata,
            self.cover_format,
        )
        download_item.cover_url = self.interface.get_cover_url(
            download_item.cover_url_template,
            self.cover_size,
            self.cover_format,
        )

        cover_file_extension = await self.interface.get_cover_file_extension(
            download_item.cover_url,
            self.cover_format,
        )
        if cover_file_extension:
            download_item.cover_path = self.get_cover_path(
                download_item.final_path,
                cover_file_extension,
            )

        return download_item

    async def download(
        self,
        download_item: DownloadItem,
    ) -> None:
        await self.download_ytdlp(
            download_item.stream_info.video_track.stream_url,
            download_item.staged_path,
        )

        cover_bytes = (
            await self.interface.get_cover_bytes(download_item.cover_url)
            if self.cover_format != CoverFormat.RAW
            else None
        )
        await self.apply_tags(
            download_item.staged_path,
            download_item.media_tags,
            cover_bytes,
        )



import logging

from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from ..interface.enums import UploadedVideoQuality
from ..interface.types import MediaTags
from .constants import UPLOADED_VIDEO_QUALITY_RANK
from .interface import AppleMusicInterface
from .types import MediaFileFormat, StreamInfo, StreamInfoAv

logger = logging.getLogger(__name__)


class AppleMusicUploadedVideoInterface(AppleMusicInterface):
    def __init__(self, interface: AppleMusicInterface):
        self.__dict__.update(interface.__dict__)

    def get_stream_url_best(self, metadata: dict) -> str:
        best_quality = next(
            (
                quality
                for quality in UPLOADED_VIDEO_QUALITY_RANK
                if metadata["attributes"]["assetTokens"].get(quality)
            ),
            None,
        )
        return metadata["attributes"]["assetTokens"][best_quality]

    async def get_stream_url_from_user(self, metadata: dict) -> str:
        qualities = list(metadata["attributes"]["assetTokens"].keys())
        choices = [
            Choice(
                name=quality,
                value=quality,
            )
            for quality in qualities
        ]
        selected = await inquirer.select(
            message="Select which quality to download:",
            choices=choices,
        ).execute_async()

        return metadata["attributes"]["assetTokens"][selected]

    async def get_stream_url(
        self, metadata: dict, quality: UploadedVideoQuality
    ) -> str:
        if quality == UploadedVideoQuality.BEST:
            stream_url = self.get_stream_url_best(metadata)

        if quality == UploadedVideoQuality.ASK:
            stream_url = await self.get_stream_url_from_user(metadata)

        logger.debug(f"Stream URL: {stream_url}")

        return stream_url

    async def get_stream_info(
        self,
        metadata: dict,
        quality: UploadedVideoQuality,
    ) -> StreamInfo:
        stream_url = await self.get_stream_url(metadata, quality)
        stream_info = StreamInfoAv(
            file_format=MediaFileFormat.M4V,
            video_track=StreamInfo(
                stream_url=stream_url,
            ),
        )
        return stream_info

    def get_tags(self, metadata: dict) -> MediaTags:
        attributes = metadata["attributes"]
        upload_date = attributes.get("uploadDate")

        tags = MediaTags(
            artist=attributes.get("artistName"),
            date=self.parse_date(upload_date) if upload_date else None,
            title=attributes.get("name"),
            title_id=int(metadata["id"]),
            storefront=int(self.itunes_api.storefront_id.split("-")[0]),
        )
        logger.debug(f"Tags: {tags}")

        return tags



import asyncio
import logging
from functools import wraps
from pathlib import Path

import click
import colorama
from dataclass_click import dataclass_click
from httpx import ConnectError

from .. import __version__
from ..api import AppleMusicApi, ItunesApi
from ..downloader import (
    AppleMusicBaseDownloader,
    AppleMusicDownloader,
    AppleMusicMusicVideoDownloader,
    AppleMusicSongDownloader,
    AppleMusicUploadedVideoDownloader,
    DownloadItem,
    DownloadMode,
    GamdlError,
    RemuxMode,
)
from ..interface import (
    AppleMusicInterface,
    AppleMusicMusicVideoInterface,
    AppleMusicSongInterface,
    AppleMusicUploadedVideoInterface,
    SongCodec,
)
from .cli_config import CliConfig
from .config_file import ConfigFile
from .constants import X_NOT_IN_PATH
from .utils import CustomLoggerFormatter, prompt_path

logger = logging.getLogger(__name__)


def make_sync(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper


@click.command()
@click.help_option("-h", "--help")
@click.version_option(__version__, "-v", "--version")
@dataclass_click(CliConfig)
@ConfigFile.loader
@make_sync
async def main(config: CliConfig):
    colorama.just_fix_windows_console()

    root_logger = logging.getLogger(__name__.split(".")[0])
    root_logger.setLevel(config.log_level)
    root_logger.propagate = False

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(CustomLoggerFormatter())
    root_logger.addHandler(stream_handler)

    if config.log_file:
        file_handler = logging.FileHandler(config.log_file, encoding="utf-8")
        file_handler.setFormatter(CustomLoggerFormatter(use_colors=False))
        root_logger.addHandler(file_handler)

    logger.info(f"Starting Gamdl {__version__}")

    if config.use_wrapper:
        try:
            apple_music_api = await AppleMusicApi.create_from_wrapper(
                wrapper_account_url=config.wrapper_account_url,
                language=config.language,
            )
        except ConnectError:
            logger.critical(
                "Could not connect to the wrapper account API. "
                "Make sure the wrapper is running and the URL is correct."
            )
            return
    else:
        cookies_path = prompt_path(config.cookies_path)
        apple_music_api = await AppleMusicApi.create_from_netscape_cookies(
            cookies_path=cookies_path,
            language=config.language,
        )

    itunes_api = ItunesApi(
        apple_music_api.storefront,
        apple_music_api.language,
    )

    if not apple_music_api.active_subscription:
        logger.critical(
            "No active Apple Music subscription found, you won't be able to download"
            " anything"
        )
        return
    if apple_music_api.account_restrictions:
        logger.warning(
            "Your account has content restrictions enabled, some content may not be"
            " downloadable"
        )

    interface = AppleMusicInterface(
        apple_music_api,
        itunes_api,
    )
    song_interface = AppleMusicSongInterface(interface)
    music_video_interface = AppleMusicMusicVideoInterface(interface)
    uploaded_video_interface = AppleMusicUploadedVideoInterface(interface)

    base_downloader = AppleMusicBaseDownloader(
        output_path=config.output_path,
        temp_path=config.temp_path,
        wvd_path=config.wvd_path,
        overwrite=config.overwrite,
        save_cover=config.save_cover,
        save_playlist=config.save_playlist,
        nm3u8dlre_path=config.nm3u8dlre_path,
        mp4decrypt_path=config.mp4decrypt_path,
        ffmpeg_path=config.ffmpeg_path,
        mp4box_path=config.mp4box_path,
        use_wrapper=config.use_wrapper,
        wrapper_decrypt_ip=config.wrapper_decrypt_ip,
        download_mode=config.download_mode,
        cover_format=config.cover_format,
        album_folder_template=config.album_folder_template,
        compilation_folder_template=config.compilation_folder_template,
        no_album_folder_template=config.no_album_folder_template,
        single_disc_file_template=config.single_disc_file_template,
        multi_disc_file_template=config.multi_disc_file_template,
        no_album_file_template=config.no_album_file_template,
        playlist_file_template=config.playlist_file_template,
        date_tag_template=config.date_tag_template,
        exclude_tags=config.exclude_tags,
        cover_size=config.cover_size,
        truncate=config.truncate,
    )
    song_downloader = AppleMusicSongDownloader(
        base_downloader=base_downloader,
        interface=song_interface,
        codec_priority=config.song_codec_piority,
        synced_lyrics_format=config.synced_lyrics_format,
        no_synced_lyrics=config.no_synced_lyrics,
        synced_lyrics_only=config.synced_lyrics_only,
        use_album_date=config.use_album_date,
        fetch_extra_tags=config.fetch_extra_tags,
    )
    music_video_downloader = AppleMusicMusicVideoDownloader(
        base_downloader=base_downloader,
        interface=music_video_interface,
        codec_priority=config.music_video_codec_priority,
        remux_mode=config.music_video_remux_mode,
        remux_format=config.music_video_remux_format,
        resolution=config.music_video_resolution,
    )
    uploaded_video_downloader = AppleMusicUploadedVideoDownloader(
        base_downloader=base_downloader,
        interface=uploaded_video_interface,
        quality=config.uploaded_video_quality,
    )
    downloader = AppleMusicDownloader(
        interface=interface,
        base_downloader=base_downloader,
        song_downloader=song_downloader,
        music_video_downloader=music_video_downloader,
        uploaded_video_downloader=uploaded_video_downloader,
        artist_auto_select=config.artist_auto_select,
    )

    if not config.synced_lyrics_only:
        if (
            config.download_mode == DownloadMode.NM3U8DLRE
            and not base_downloader.full_nm3u8dlre_path
        ):
            logger.critical(X_NOT_IN_PATH.format("N_m3u8DL-RE", config.nm3u8dlre_path))
            return

        missing_music_video_paths = []

        if not base_downloader.full_ffmpeg_path and (
            config.music_video_remux_mode == RemuxMode.FFMPEG
            or config.download_mode == DownloadMode.NM3U8DLRE
        ):
            missing_music_video_paths.append(
                X_NOT_IN_PATH.format("ffmpeg", config.ffmpeg_path)
            )

        if (
            not base_downloader.full_mp4box_path
            and config.music_video_remux_mode == RemuxMode.MP4BOX
        ):
            missing_music_video_paths.append(
                X_NOT_IN_PATH.format("MP4Box", config.mp4box_path)
            )

        if not base_downloader.full_mp4decrypt_path and (
            config.song_codec_piority
            not in (SongCodec.AAC_LEGACY, SongCodec.AAC_HE_LEGACY)
            or config.music_video_remux_mode == RemuxMode.MP4BOX
        ):
            missing_music_video_paths.append(
                X_NOT_IN_PATH.format("mp4decrypt", config.mp4decrypt_path)
            )

        if missing_music_video_paths:
            logger.warning(
                "Music videos will not be downloaded due to missing dependencies:\n"
                + "\n".join(missing_music_video_paths)
            )

        if (
            any(not codec.is_legacy() for codec in config.song_codec_piority)
            and not config.use_wrapper
        ):
            logger.warning(
                "You have chosen an experimental song codec "
                "without enabling wrapper. "
                "They're not guaranteed to work due to API limitations."
            )

    if config.read_urls_as_txt:
        urls_from_file = []
        for url in config.urls:
            if Path(url).is_file() and Path(url).exists():
                urls_from_file.extend(
                    [
                        line.strip()
                        for line in Path(url).read_text(encoding="utf-8").splitlines()
                        if line.strip()
                    ]
                )
        urls = urls_from_file
    else:
        urls = config.urls

    error_count = 0
    for url_index, url in enumerate(urls, 1):
        url_progress = click.style(f"[URL {url_index}/{len(urls)}]", dim=True)
        logger.info(url_progress + f' Processing "{url}"')
        download_queue = None
        try:
            url_info = downloader.get_url_info(url)
            if not url_info:
                logger.warning(
                    url_progress + f' Could not parse "{url}", skipping.',
                )
                continue

            download_queue = await downloader.get_download_queue(url_info)
            if not download_queue:
                logger.warning(
                    url_progress
                    + f' No downloadable media found for "{url}", skipping.',
                )
                continue
        except KeyboardInterrupt:
            exit(1)
        except Exception as e:
            error_count += 1
            logger.error(
                url_progress + f' Error processing "{url}"',
                exc_info=not config.no_exceptions,
            )

        if not download_queue:
            continue

        for download_index, download_item in enumerate(
            download_queue,
            1,
        ):
            download_queue_progress = click.style(
                f"[Track {download_index}/{len(download_queue)}]",
                dim=True,
            )
            media_title = (
                download_item.media_metadata["attributes"]["name"]
                if isinstance(
                    download_item,
                    DownloadItem,
                )
                else "Unknown Title"
            )
            logger.info(download_queue_progress + f' Downloading "{media_title}"')

            try:
                await downloader.download(download_item)
            except GamdlError as e:
                logger.warning(
                    download_queue_progress + f' Skipping "{media_title}": {e}'
                )
                continue
            except KeyboardInterrupt:
                exit(1)
            except Exception as e:
                error_count += 1
                logger.error(
                    download_queue_progress + f' Error downloading "{media_title}"',
                    exc_info=not config.no_exceptions,
                )

    logger.info(f"Finished with {error_count} error(s)")



import logging

import httpx

from ..utils import safe_json
from .constants import ITUNES_LOOKUP_API_URL, ITUNES_PAGE_API_URL, STOREFRONT_IDS
from .exceptions import ApiError

logger = logging.getLogger(__name__)


class ItunesApi:
    def __init__(
        self,
        storefront: str = "us",
        language: str = "en-US",
    ) -> None:
        self.storefront = storefront
        self.language = language
        self.initialize()

    def initialize(self) -> None:
        self._initialize_storefront_id()
        self._initialize_client()

    def _initialize_storefront_id(self) -> None:
        try:
            self.storefront_id = STOREFRONT_IDS[self.storefront.upper()]
        except KeyError:
            raise Exception(f"No storefront id for {self.storefront}")

    def _initialize_client(self) -> None:
        self.client = httpx.AsyncClient(
            params={
                "country": self.storefront,
                "lang": self.language,
            },
            headers={
                "X-Apple-Store-Front": f"{self.storefront_id} t:music31",
            },
            timeout=60.0,
        )

    async def get_lookup_result(
        self,
        media_id: str,
        entity: str = "album",
    ) -> dict:
        response = await self.client.get(
            ITUNES_LOOKUP_API_URL,
            params={
                "id": media_id,
                "entity": entity,
            },
        )
        lookup_result = safe_json(response)

        if response.status_code != 200 or lookup_result is None:
            raise ApiError(
                message=response.text,
                status_code=response.status_code,
            )

        logger.debug(f"Lookup result: {lookup_result}")

        return lookup_result

    async def get_itunes_page(
        self,
        media_type: str,
        media_id: str,
    ) -> dict:
        response = await self.client.get(
            f"{ITUNES_PAGE_API_URL}/{media_type}/{media_id}"
        )
        itunes_page = safe_json(response)

        if response.status_code != 200 or itunes_page is None:
            raise ApiError(
                message=response.text,
                status_code=response.status_code,
            )

        logger.debug(f"iTunes page: {itunes_page}")

        return itunes_page



from ..utils import GamdlError


class MediaFileExists(GamdlError):
    def __init__(self, media_path: str):
        super().__init__(f"Media file already exists at path: {media_path}")


class NotStreamable(GamdlError):
    def __init__(self, media_id: str):
        super().__init__(f"Media ID is not streamable: {media_id}")


class FormatNotAvailable(GamdlError):
    def __init__(self, media_id: str):
        super().__init__(f"Requested format is not available for media ID: {media_id}")


class ExecutableNotFound(GamdlError):
    def __init__(self, executable: str):
        super().__init__(f"Executable not found: {executable}")


class SyncedLyricsOnly(GamdlError):
    def __init__(self):
        super().__init__("Only downloading synced lyrics is supported")


class UnsupportedMediaType(GamdlError):
    def __init__(self, media_type: str):
        super().__init__(f"Unsupported media type: {media_type}")



MEDIA_TYPE_STR_MAP = {
    1: "Song",
    6: "Music Video",
}

MEDIA_RATING_STR_MAP = {
    0: "None",
    1: "Explicit",
    2: "Clean",
}

LEGACY_SONG_CODECS = {"aac-legacy", "aac-he-legacy"}

DRM_DEFAULT_KEY_MAPPING = {
    "urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed": (
        "data:text/plain;base64,AAAAOHBzc2gAAAAA7e+LqXnWSs6jyCfc1R0h7QAAABgSEAAAAAA"
        "AAAAAczEvZTEgICBI88aJmwY="
    ),
    "com.microsoft.playready": (
        "data:text/plain;charset=UTF-16;base64,vgEAAAEAAQC0ATwAVwBSAE0ASABFAEEARABF"
        "AFIAIAB4AG0AbABuAHMAPQAiAGgAdAB0AHAAOgAvAC8AcwBjAGgAZQBtAGEAcwAuAG0AaQBjAH"
        "IAbwBzAG8AZgB0AC4AYwBvAG0ALwBEAFIATQAvADIAMAAwADcALwAwADMALwBQAGwAYQB5AFIA"
        "ZQBhAGQAeQBIAGUAYQBkAGUAcgAiACAAdgBlAHIAcwBpAG8AbgA9ACIANAAuADMALgAwAC4AMA"
        "AiAD4APABEAEEAVABBAD4APABQAFIATwBUAEUAQwBUAEkATgBGAE8APgA8AEsASQBEAFMAPgA8"
        "AEsASQBEACAAQQBMAEcASQBEAD0AIgBBAEUAUwBDAEIAQwAiACAAVgBBAEwAVQBFAD0AIgBBAE"
        "EAQQBBAEEAQQBBAEEAQQBBAEIAegBNAFMAOQBsAE0AUwBBAGcASQBBAD0APQAiAD4APAAvAEsA"
        "SQBEAD4APAAvAEsASQBEAFMAPgA8AC8AUABSAE8AVABFAEMAVABJAE4ARgBPAD4APAAvAEQAQQ"
        "BUAEEAPgA8AC8AVwBSAE0ASABFAEEARABFAFIAPgA="
    ),
    "com.apple.streamingkeydelivery": "skd://itunes.apple.com/P000000000/s1/e1",
}
MP4_FORMAT_CODECS = ["ec-3", "hvc1", "audio-atmos", "audio-ec3"]
SONG_CODEC_REGEX_MAP = {
    "aac": r"audio-stereo-\d+",
    "aac-he": r"audio-HE-stereo-\d+",
    "aac-binaural": r"audio-stereo-\d+-binaural",
    "aac-downmix": r"audio-stereo-\d+-downmix",
    "aac-he-binaural": r"audio-HE-stereo-\d+-binaural",
    "aac-he-downmix": r"audio-HE-stereo-\d+-downmix",
    "atmos": r"audio-atmos-.*",
    "ac3": r"audio-ac3-.*",
    "alac": r"audio-alac-.*",
}

FOURCC_MAP = {
    "h264": "avc1",
    "h265": "hvc1",
}

UPLOADED_VIDEO_QUALITY_RANK = [
    "1080pHdVideo",
    "720pHdVideo",
    "sdVideoWithPlusAudio",
    "sdVideo",
    "sd480pVideo",
    "provisionalUploadVideo",
]

IMAGE_FILE_EXTENSION_MAP = {
    "jpeg": ".jpg",
    "tiff": ".tif",
}



from ..utils import GamdlError


class ApiError(GamdlError):
    def __init__(self, message: str, status_code: int):
        super().__init__(f"API Error {status_code}: {message}")
        self.status_code = status_code



import logging
import re
import typing
from http.cookiejar import MozillaCookieJar
from urllib.parse import parse_qs, urlparse

import httpx

from ..utils import get_response, raise_for_status, safe_json
from .constants import (
    AMP_API_URL,
    APPLE_MUSIC_COOKIE_DOMAIN,
    APPLE_MUSIC_HOMEPAGE_URL,
    LICENSE_API_URL,
    WEBPLAYBACK_API_URL,
)
from .exceptions import ApiError

logger = logging.getLogger(__name__)


class AppleMusicApi:
    def __init__(
        self,
        storefront: str = "us",
        language: str = "en-US",
        media_user_token: str | None = None,
        developer_token: str | None = None,
    ) -> None:
        self.storefront = storefront
        self.language = language
        self.media_user_token = media_user_token
        self.token = developer_token

    @classmethod
    async def create_from_netscape_cookies(
        cls,
        cookies_path: str = "./cookies.txt",
        *args,
        **kwargs,
    ) -> "AppleMusicApi":
        cookies = MozillaCookieJar(cookies_path)
        cookies.load(ignore_discard=True, ignore_expires=True)
        parse_cookie = lambda name: next(
            (
                cookie.value
                for cookie in cookies
                if cookie.name == name and cookie.domain == APPLE_MUSIC_COOKIE_DOMAIN
            ),
            None,
        )

        media_user_token = parse_cookie("media-user-token")
        if not media_user_token:
            raise ValueError(
                '"media-user-token" cookie not found in cookies. '
                "Make sure you have exported the cookies from the Apple Music webpage "
                "and are logged in with an active subscription."
            )

        return await cls.create(
            storefront=None,
            media_user_token=media_user_token,
            developer_token=None,
            *args,
            **kwargs,
        )

    @classmethod
    async def create_from_wrapper(
        cls,
        wrapper_account_url: str = "http://127.0.0.1:30020/",
        *args,
        **kwargs,
    ) -> "AppleMusicApi":
        wrapper_account_response = await get_response(wrapper_account_url)
        wrapper_account_info = safe_json(wrapper_account_response)

        return await cls.create(
            storefront=None,
            media_user_token=wrapper_account_info["music_token"],
            developer_token=wrapper_account_info["dev_token"],
            *args,
            **kwargs,
        )

    @classmethod
    async def create(
        cls,
        storefront: str | None = "us",
        language: str = "en-US",
        media_user_token: str | None = None,
        developer_token: str | None = None,
    ) -> "AppleMusicApi":
        api = cls(
            storefront=storefront,
            language=language,
            media_user_token=media_user_token,
            developer_token=developer_token,
        )
        await api.initialize()
        return api

    async def initialize(self) -> None:
        await self._initialize_client()
        await self._initialize_token()
        await self._initialize_account_info()

    async def _initialize_client(self) -> None:
        self.client = httpx.AsyncClient(
            headers={
                "accept": "*/*",
                "accept-language": "en-US",
                "origin": APPLE_MUSIC_HOMEPAGE_URL,
                "priority": "u=1, i",
                "referer": APPLE_MUSIC_HOMEPAGE_URL,
                "sec-ch-ua": '"Google Chrome";v="137", "Chromium";v="137", "Not/A)Brand";v="24"',
                "sec-ch-ua-mobile": "?0",
                "sec-ch-ua-platform": '"Windows"',
                "sec-fetch-dest": "empty",
                "sec-fetch-mode": "cors",
                "sec-fetch-site": "same-site",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36",
            },
            params={
                "l": self.language,
            },
            follow_redirects=True,
            timeout=60.0,
        )

    async def _get_token(self) -> str:
        response = await self.client.get(APPLE_MUSIC_HOMEPAGE_URL)
        home_page = response.text

        index_js_uri_match = re.search(
            r"/(assets/index-legacy[~-][^/\"]+\.js)",
            home_page,
        )
        if not index_js_uri_match:
            raise Exception("index.js URI not found in Apple Music homepage")
        index_js_uri = index_js_uri_match.group(1)

        response = await self.client.get(f"{APPLE_MUSIC_HOMEPAGE_URL}/{index_js_uri}")
        index_js_page = response.text

        token_match = re.search('(?=eyJh)(.*?)(?=")', index_js_page)
        if not token_match:
            raise Exception("Token not found in index.js page")
        token = token_match.group(1)

        logger.debug(f"Token: {token}")
        return token

    async def _initialize_token(self) -> None:
        self.token = self.token or await self._get_token()
        self.client.headers.update({"authorization": f"Bearer {self.token}"})

    async def _initialize_account_info(self) -> None:
        if not self.media_user_token:
            return

        self.client.cookies.update(
            {
                "media-user-token": self.media_user_token,
            }
        )

        self.account_info = await self.get_account_info()
        self.storefront = self.account_info["meta"]["subscription"]["storefront"]

    @property
    def active_subscription(self) -> bool:
        return (
            getattr(self, "account_info", {})
            .get("meta", {})
            .get("subscription", {})
            .get("active", False)
        )

    @property
    def account_restrictions(self) -> dict | None:
        data = getattr(self, "account_info", {}).get("data", [])
        if not data:
            return None
        return data[0].get("attributes", {}).get("restrictions")

    async def get_account_info(self, meta: str = "subscription") -> dict:
        account_info = await self._amp_request(
            f"/v1/me/account",
            {
                "meta": meta,
            },
        )
        logger.debug(f"Account info: {account_info}")

        return account_info

    async def _amp_request(
        self,
        endpoint: str,
        params: dict | None = None,
    ) -> dict:
        response = await self.client.get(
            AMP_API_URL + endpoint,
            params=params or {},
        )
        response_json = safe_json(response)

        if (
            response.status_code != 200
            or response_json is None
            or "errors" in response_json
        ):
            raise ApiError(
                message=response.text,
                status_code=response.status_code,
            )

        return response_json

    async def get_song(
        self,
        song_id: str,
        extend: str = "extendedAssetUrls",
        include: str = "lyrics,albums",
    ) -> dict | None:
        song = await self._amp_request(
            f"/v1/catalog/{self.storefront}/songs/{song_id}",
            {
                "extend": extend,
                "include": include,
            },
        )
        logger.debug(f"Song: {song}")

        return song

    async def get_music_video(
        self,
        music_video_id: str,
        include: str = "albums",
    ) -> dict | None:
        music_video = await self._amp_request(
            f"/v1/catalog/{self.storefront}/music-videos/{music_video_id}",
            {
                "include": include,
            },
        )
        logger.debug(f"Music video: {music_video}")

        return music_video

    async def get_uploaded_video(
        self,
        post_id: str,
    ) -> dict | None:
        uploaded_video = await self._amp_request(
            f"/v1/catalog/{self.storefront}/uploaded-videos/{post_id}",
        )
        logger.debug(f"Uploaded video: {uploaded_video}")

        return uploaded_video

    async def get_album(
        self,
        album_id: str,
        extend: str = "extendedAssetUrls",
    ) -> dict | None:
        album = await self._amp_request(
            f"/v1/catalog/{self.storefront}/albums/{album_id}",
            {
                "extend": extend,
            },
        )
        logger.debug(f"Album: {album}")

        return album

    async def get_playlist(
        self,
        playlist_id: str,
        limit_tracks: int = 300,
        extend: str = "extendedAssetUrls",
    ) -> dict | None:
        playlist = await self._amp_request(
            f"/v1/catalog/{self.storefront}/playlists/{playlist_id}",
            {
                "limit[tracks]": limit_tracks,
                "extend": extend,
            },
        )
        logger.debug(f"Playlist: {playlist}")

        return playlist

    async def get_artist(
        self,
        artist_id: str,
        include: str = "albums,music-videos",
        views: str = "full-albums,compilation-albums,live-albums,singles,top-songs",
        limit: int = 100,
    ) -> dict | None:
        artist = await self._amp_request(
            f"/v1/catalog/{self.storefront}/artists/{artist_id}",
            {
                "include": include,
                "views": views,
                **{
                    f"limit[{_include}]": limit
                    for _include in [*include.split(","), *views.split(",")]
                },
            },
        )
        logger.debug(f"Artist: {artist}")

        return artist

    async def get_library_album(
        self,
        album_id: str,
        extend: str = "extendedAssetUrls",
    ) -> dict | None:
        album = await self._amp_request(
            f"/v1/me/library/albums/{album_id}",
            {
                "extend": extend,
            },
        )
        logger.debug(f"Library album: {album}")

        return album

    async def get_library_playlist(
        self,
        playlist_id: str,
        include: str = "tracks",
        limit: int = 100,
        extend: str = "extendedAssetUrls",
    ) -> dict | None:
        playlist = await self._amp_request(
            f"/v1/me/library/playlists/{playlist_id}",
            {
                "include": include,
                **{f"limit[{_include}]": limit for _include in include.split(",")},
                "extend": extend,
            },
        )
        logger.debug(f"Library playlist: {playlist}")

        return playlist

    async def get_search_results(
        self,
        term: str,
        types: str = "songs,music-videos,albums,playlists,artists",
        limit: int = 50,
        offset: int = 0,
    ) -> dict:
        search_results = await self._amp_request(
            f"/v1/catalog/{self.storefront}/search",
            {
                "term": term,
                "types": types,
                "limit": limit,
                "offset": offset,
            },
        )
        logger.debug(f"Search results: {search_results}")

        return search_results

    async def extend_api_data(
        self,
        api_response: dict,
        extend: str = "extendedAssetUrls",
    ) -> typing.AsyncGenerator[dict, None]:
        next_uri = api_response.get("next")
        if not next_uri:
            return

        next_uri_params = parse_qs(urlparse(next_uri).query)
        limit = int(next_uri_params["offset"][0])
        while next_uri:
            extended_api_data = await self._get_extended_api_data(
                next_uri,
                limit,
                extend,
            )
            yield extended_api_data
            next_uri = extended_api_data.get("next")

    async def _get_extended_api_data(
        self,
        next_uri: str,
        limit: int,
        extend: str,
    ) -> dict:
        next_uri_params = parse_qs(urlparse(next_uri).query)
        params = {
            "limit": limit,
            "offset": next_uri_params["offset"][0],
            "extend": extend,
        }
        extended_api_data = await self._amp_request(next_uri, params)
        logger.debug(f"Extended API data: {extended_api_data}")

        return extended_api_data

    async def get_webplayback(
        self,
        track_id: str,
    ) -> dict:
        response = await self.client.post(
            WEBPLAYBACK_API_URL,
            json={
                "salableAdamId": track_id,
                "language": self.language,
            },
        )
        webplayback = safe_json(response)

        if (
            response.status_code != 200
            or webplayback is None
            or "dialog" in webplayback
        ):
            raise ApiError(
                message=response.text,
                status_code=response.status_code,
            )

        return webplayback

    async def get_license_exchange(
        self,
        track_id: str,
        track_uri: str,
        challenge: str,
        key_system: str = "com.widevine.alpha",
    ) -> dict:
        response = await self.client.post(
            LICENSE_API_URL,
            json={
                "challenge": challenge,
                "key-system": key_system,
                "uri": track_uri,
                "adamId": track_id,
                "isLibrary": False,
                "user-initiated": True,
            },
        )
        license_exchange = safe_json(response)

        if (
            response.status_code != 200
            or license_exchange is None
            or license_exchange.get("status") != 0
        ):
            raise ApiError(
                message=response.text,
                status_code=response.status_code,
            )

        logger.debug(f"License exchange: {license_exchange}")

        return license_exchange



import configparser
import typing
from functools import wraps
from pathlib import Path

import click
import click.types as click_types

from .cli_config import CliConfig
from .constants import EXCLUDED_CONFIG_FILE_PARAMS
from .utils import Csv


class ConfigFile:
    def __init__(
        self,
        config_path: str,
        section_name: str = "gamdl",
    ) -> None:
        self.config_path = config_path
        self.section_name = section_name

        self.click_context = click.get_current_context()
        self._read_config_file()

    def _read_config_file(self) -> None:
        self.config = configparser.ConfigParser(interpolation=None)

        if Path(self.config_path).exists():
            self.config.read(self.config_path, encoding="utf-8")
        else:
            Path(self.config_path).parent.mkdir(parents=True, exist_ok=True)

        if not self.config.has_section(self.section_name):
            self.config.add_section(self.section_name)

    def _write_config_file(self) -> None:
        with open(self.config_path, "w", encoding="utf-8") as config_file:
            self.config.write(config_file)

    def _serialize_param_default(self, param: click.Parameter) -> str:
        if param.default is None:
            return "null"

        if isinstance(param.type, Csv):
            return ",".join(
                item.value if hasattr(item, "value") else str(item)
                for item in param.default
            )

        if isinstance(param.type, click_types.FuncParamType):
            return param.default.value

        if isinstance(param.type, click_types.BoolParamType):
            return "true" if param.default else "false"

        if isinstance(
            param.type,
            click_types.Choice
            | click_types.Path
            | click_types.StringParamType
            | click_types.IntParamType,
        ):
            return str(param.default)

        raise NotImplementedError(
            f"Serialization for parameter '{param.name}' of type "
            f"'{type(param.type)}' is not implemented."
        )

    def _add_param_default_to_config(
        self,
        param: click.Parameter,
    ) -> bool:
        if self.config.has_option(self.section_name, param.name):
            return False

        value = self._serialize_param_default(param)
        self.config.set(self.section_name, param.name, value)

        return True

    def _parse_param_from_config(
        self,
        param: click.Parameter,
    ) -> typing.Any:
        value = self.config[self.section_name].get(param.name)
        if value is None:
            return param.default

        if value == "null":
            return None

        if not isinstance(param.type, click_types.ParamType):
            raise NotImplementedError(
                f"Parsing for parameter '{param.name}' of type "
                f"'{type(param.type)}' is not implemented."
            )

        return param.type.convert(value, None, None)

    def add_params_default_to_config(self) -> None:
        has_changes = False

        for param in self.click_context.command.params:
            if param.name in EXCLUDED_CONFIG_FILE_PARAMS:
                continue

            has_changes = self._add_param_default_to_config(param) or has_changes

        if has_changes:
            self._write_config_file()

    def cleanup_unknown_params(self) -> None:
        param_names = {info.name for info in self.click_context.command.params}
        has_changes = False

        for key in list(self.config[self.section_name].keys()):
            if key not in param_names:
                self.config.remove_option(self.section_name, key)
                has_changes = True

        if has_changes:
            self._write_config_file()

    def update_params_from_config(self) -> None:
        for param in self.click_context.command.params:
            if (
                self.click_context.get_parameter_source(param.name)
                == click.core.ParameterSource.COMMANDLINE
            ):
                continue

            if self.config.has_option(self.section_name, param.name):
                self.click_context.params[param.name] = self._parse_param_from_config(
                    param
                )

    def get_cli_config(self) -> CliConfig:
        config_dict = {}
        for param in self.click_context.command.params:
            if param.name in {"help", "version"}:
                continue

            config_dict[param.name] = self.click_context.params.get(
                param.name, param.default
            )
        return CliConfig(**config_dict)

    def load(self) -> CliConfig:
        self.cleanup_unknown_params()
        self.add_params_default_to_config()
        self.update_params_from_config()
        return self.get_cli_config()

    @staticmethod
    def loader(func):
        @wraps(func)
        def wrapper(cli_config: CliConfig):
            ctx = click.get_current_context()
            config_path = ctx.params.get("config_path")
            no_config_file = ctx.params.get("no_config_file")
            if config_path and not no_config_file:
                cli_config = ConfigFile(config_path).load()
            return func(cli_config)

        return wrapper



from pathlib import Path

from ..interface.enums import CoverFormat, SongCodec, SyncedLyricsFormat
from ..interface.interface_song import AppleMusicSongInterface
from ..interface.types import DecryptionKeyAv
from .amdecrypt import decrypt_file, decrypt_file_hex
from .downloader_base import AppleMusicBaseDownloader
from .types import DownloadItem


class AppleMusicSongDownloader(AppleMusicBaseDownloader):
    def __init__(
        self,
        base_downloader: AppleMusicBaseDownloader,
        interface: AppleMusicSongInterface,
        codec_priority: SongCodec = [SongCodec.AAC_LEGACY],
        synced_lyrics_format: SyncedLyricsFormat = SyncedLyricsFormat.LRC,
        no_synced_lyrics: bool = False,
        synced_lyrics_only: bool = False,
        use_album_date: bool = False,
        fetch_extra_tags: bool = False,
    ):
        self.__dict__.update(base_downloader.__dict__)
        self.interface = interface
        self.codec_priority = codec_priority
        self.synced_lyrics_format = synced_lyrics_format
        self.no_synced_lyrics = no_synced_lyrics
        self.synced_lyrics_only = synced_lyrics_only
        self.use_album_date = use_album_date
        self.fetch_extra_tags = fetch_extra_tags

    async def get_download_item(
        self,
        song_metadata: dict,
        playlist_metadata: dict = None,
    ) -> DownloadItem:
        download_item = DownloadItem()

        download_item.media_metadata = song_metadata
        download_item.playlist_metadata = playlist_metadata

        song_id = self.interface.get_media_id_of_library_media(song_metadata)

        download_item.lyrics = await self.interface.get_lyrics(
            song_metadata,
            self.synced_lyrics_format,
        )

        webplayback = await self.interface.apple_music_api.get_webplayback(song_id)
        download_item.media_tags = await self.interface.get_tags(
            webplayback,
            download_item.lyrics.unsynced if download_item.lyrics else None,
            self.use_album_date,
        )
        if self.fetch_extra_tags:
            download_item.extra_tags = await self.interface.get_extra_tags(
                song_metadata,
            )

        if playlist_metadata:
            download_item.playlist_tags = self.get_playlist_tags(
                playlist_metadata,
                song_metadata,
            )
            download_item.playlist_file_path = self.get_playlist_file_path(
                download_item.playlist_tags,
            )

        download_item.final_path = self.get_final_path(
            download_item.media_tags,
            ".m4a",
            download_item.playlist_tags,
        )
        download_item.synced_lyrics_path = self.get_lyrics_synced_path(
            download_item.final_path,
        )

        if self.synced_lyrics_only:
            return download_item

        for codec in self.codec_priority:
            download_item.stream_info = await self.interface.get_stream_info(
                codec,
                song_metadata,
                webplayback,
            )
            if download_item.stream_info:
                break

        if download_item.stream_info.audio_track.legacy:
            download_item.decryption_key = (
                await self.interface.get_decryption_key_legacy(
                    download_item.stream_info,
                    self.cdm,
                )
            )
        elif (
            not self.use_wrapper
            and download_item.stream_info
            and download_item.stream_info.audio_track.widevine_pssh
        ):
            download_item.decryption_key = await self.interface.get_decryption_key(
                download_item.stream_info,
                self.cdm,
            )

        download_item.cover_url_template = self.interface.get_cover_url_template(
            song_metadata,
            self.cover_format,
        )
        download_item.cover_url = self.interface.get_cover_url(
            download_item.cover_url_template,
            self.cover_size,
            self.cover_format,
        )

        download_item.random_uuid = self.get_random_uuid()
        if download_item.stream_info and download_item.stream_info.file_format:
            download_item.staged_path = self.get_temp_path(
                song_id,
                download_item.random_uuid,
                "staged",
                "." + download_item.stream_info.file_format.value,
            )
        else:
            download_item.staged_path = None

        cover_file_extension = await self.interface.get_cover_file_extension(
            download_item.cover_url,
            self.cover_format,
        )
        if cover_file_extension:
            download_item.cover_path = self.get_cover_path(
                download_item.final_path,
                cover_file_extension,
            )

        return download_item

    async def decrypt_amdecrypt(
        self,
        input_path: str,
        output_path: str,
        media_id: str,
        fairplay_key: str,
    ) -> None:
        await decrypt_file(
            self.wrapper_decrypt_ip,
            media_id,
            fairplay_key,
            input_path,
            output_path,
        )

    async def decrypt_amdecrypt_hex(
        self,
        input_path: str,
        output_path: str,
        decryption_key: str,
        legacy: bool = False,
    ) -> None:
        await decrypt_file_hex(
            input_path,
            output_path,
            decryption_key,
            legacy=legacy,
        )

    async def stage(
        self,
        encrypted_path: str,
        staged_path: str,
        decryption_key: DecryptionKeyAv,
        legacy: bool,
        media_id: str,
        fairplay_key: str,
    ):
        if self.use_wrapper and not legacy:
            await self.decrypt_amdecrypt(
                encrypted_path,
                staged_path,
                media_id,
                fairplay_key,
            )
        else:
            await self.decrypt_amdecrypt_hex(
                encrypted_path,
                staged_path,
                decryption_key.audio_track.key,
                legacy,
            )

    def get_lyrics_synced_path(self, final_path: str) -> str:
        return str(Path(final_path).with_suffix("." + self.synced_lyrics_format.value))

    def get_cover_path(
        self,
        final_path: str,
        file_extension: str,
    ) -> str:
        return str(Path(final_path).parent / ("Cover" + file_extension))

    def write_synced_lyrics(
        self,
        synced_lyrics: str,
        lyrics_synced_path: str,
    ):
        Path(lyrics_synced_path).parent.mkdir(parents=True, exist_ok=True)
        Path(lyrics_synced_path).write_text(synced_lyrics, encoding="utf8")

    async def download(
        self,
        download_item: DownloadItem,
    ) -> None:
        if self.synced_lyrics_only:
            return

        encrypted_path = self.get_temp_path(
            download_item.media_metadata["id"],
            download_item.random_uuid,
            "encrypted",
            ".m4a",
        )
        await self.download_stream(
            download_item.stream_info.audio_track.stream_url,
            encrypted_path,
        )

        await self.stage(
            encrypted_path,
            download_item.staged_path,
            download_item.decryption_key,
            download_item.stream_info.audio_track.legacy,
            download_item.media_metadata["id"],
            download_item.stream_info.audio_track.fairplay_key,
        )

        cover_bytes = (
            await self.interface.get_cover_bytes(download_item.cover_url)
            if self.cover_format != CoverFormat.RAW
            else None
        )
        await self.apply_tags(
            download_item.staged_path,
            download_item.media_tags,
            cover_bytes,
            download_item.extra_tags,
        )



import logging
from enum import Enum
from pathlib import Path

import click


class Csv(click.ParamType):
    name = "csv"

    def __init__(
        self,
        subtype: Enum,
    ) -> None:
        self.subtype = subtype

    def convert(
        self,
        value: str,
        param: click.Parameter,
        ctx: click.Context,
    ) -> list[Enum]:
        if not isinstance(value, str):
            return value

        items = [v.strip() for v in value.split(",") if v.strip()]
        result = []

        for item in items:
            try:
                result.append(self.subtype(item))
            except ValueError as e:
                self.fail(
                    f"'{item}' is not a valid value for {self.subtype.__name__}",
                    param,
                    ctx,
                )
        return result


class CustomLoggerFormatter(logging.Formatter):
    base_format = "[%(levelname)-8s %(asctime)s]"
    format_colors = {
        logging.DEBUG: dict(dim=True),
        logging.INFO: dict(fg="green"),
        logging.WARNING: dict(fg="yellow"),
        logging.ERROR: dict(fg="red"),
        logging.CRITICAL: dict(fg="red", bold=True),
    }
    date_format = "%H:%M:%S"

    def __init__(self, use_colors: bool = True) -> None:
        super().__init__()
        self.use_colors = use_colors

    def format(self, record: logging.LogRecord) -> str:
        return logging.Formatter(
            (
                click.style(self.base_format, **self.format_colors.get(record.levelno))
                if self.use_colors
                else self.base_format
            )
            + " %(message)s",
            datefmt=self.date_format,
        ).format(record)


def prompt_path(
    input_path: str,
    is_dir: bool = False,
) -> str:
    path_validator = click.Path(
        exists=True,
        file_okay=not is_dir,
        dir_okay=is_dir,
    )
    path_type = "directory" if is_dir else "file"

    while True:
        try:
            result_path = path_validator.convert(input_path, None, None)
            break
        except click.BadParameter as e:
            input_path = click.prompt(
                (
                    f'{path_type.capitalize()} "{Path(input_path).absolute()}" does not exist. '
                    f"Create the {path_type} at the specified path, "
                    f"type a new path or drag and drop the {path_type} here. "
                    "Then, press enter to continue"
                ),
                default=input_path,
                show_default=False,
            )
            input_path = input_path.strip('"')

    return result_path



from dataclasses import dataclass
from typing import Any

from ..interface.types import (
    DecryptionKeyAv,
    Lyrics,
    MediaTags,
    PlaylistTags,
    StreamInfoAv,
)


@dataclass
class DownloadItem:
    media_metadata: dict = None
    playlist_metadata: dict = None
    random_uuid: str = None
    lyrics: Lyrics = None
    media_tags: MediaTags = None
    extra_tags: dict = None
    playlist_tags: PlaylistTags = None
    stream_info: StreamInfoAv = None
    decryption_key: DecryptionKeyAv = None
    cover_url_template: str = None
    cover_url: str = None
    staged_path: str = None
    final_path: str = None
    playlist_file_path: str = None
    synced_lyrics_path: str = None
    cover_path: str = None
    flat_filter_result: Any = None
    error: Exception = None


@dataclass
class UrlInfo:
    storefront: str = None
    type: str = None
    slug: str = None
    id: str = None
    sub_id: str = None
    library_storefront: str = None
    library_type: str = None
    library_id: str = None



import asyncio
import typing
from pathlib import Path

from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from ..api.exceptions import ApiError
from ..interface import AppleMusicInterface
from ..utils import safe_gather
from .constants import (
    ALBUM_MEDIA_TYPE,
    ARTIST_MEDIA_TYPE,
    MUSIC_VIDEO_MEDIA_TYPE,
    PLAYLIST_MEDIA_TYPE,
    SONG_MEDIA_TYPE,
    UPLOADED_VIDEO_MEDIA_TYPE,
    VALID_URL_PATTERN,
)
from .downloader_base import AppleMusicBaseDownloader
from .downloader_music_video import AppleMusicMusicVideoDownloader
from .downloader_song import AppleMusicSongDownloader
from .downloader_uploaded_video import AppleMusicUploadedVideoDownloader
from .enums import ArtistAutoSelect, DownloadMode, RemuxMode
from .exceptions import (
    ExecutableNotFound,
    FormatNotAvailable,
    MediaFileExists,
    NotStreamable,
    SyncedLyricsOnly,
    UnsupportedMediaType,
)
from .types import DownloadItem, UrlInfo


class AppleMusicDownloader:
    def __init__(
        self,
        interface: AppleMusicInterface,
        base_downloader: AppleMusicBaseDownloader,
        song_downloader: AppleMusicSongDownloader,
        music_video_downloader: AppleMusicMusicVideoDownloader,
        uploaded_video_downloader: AppleMusicUploadedVideoDownloader,
        artist_auto_select: ArtistAutoSelect | None = None,
        skip_music_videos: bool = False,
        skip_processing: bool = False,
        flat_filter: typing.Callable = None,
    ):
        self.interface = interface
        self.base_downloader = base_downloader
        self.song_downloader = song_downloader
        self.music_video_downloader = music_video_downloader
        self.uploaded_video_downloader = uploaded_video_downloader
        self.artist_auto_select = artist_auto_select
        self.skip_music_videos = skip_music_videos
        self.skip_processing = skip_processing
        self.flat_filter = flat_filter

    async def get_single_download_item(
        self,
        media_metadata: dict,
        playlist_metadata: dict = None,
    ) -> DownloadItem:
        if self.flat_filter:
            flat_filter_result = self.flat_filter(media_metadata)
            if asyncio.iscoroutine(flat_filter_result):
                flat_filter_result = await flat_filter_result

            if flat_filter_result:
                return DownloadItem(
                    media_metadata=media_metadata,
                    playlist_metadata=playlist_metadata,
                    flat_filter_result=flat_filter_result,
                )

        return await self.get_single_download_item_no_filter(
            media_metadata,
            playlist_metadata,
        )

    async def get_single_download_item_no_filter(
        self,
        media_metadata: dict,
        playlist_metadata: dict = None,
    ) -> DownloadItem:
        try:
            if not self.base_downloader.is_media_streamable(
                media_metadata,
            ):
                raise NotStreamable(media_metadata["id"])

            if media_metadata["type"] in SONG_MEDIA_TYPE:
                if not self.song_downloader:
                    raise UnsupportedMediaType(media_metadata["type"])

                download_item = await self.song_downloader.get_download_item(
                    media_metadata,
                    playlist_metadata,
                )

            if media_metadata["type"] in MUSIC_VIDEO_MEDIA_TYPE:
                if not self.music_video_downloader:
                    raise UnsupportedMediaType(media_metadata["type"])

                download_item = await self.music_video_downloader.get_download_item(
                    media_metadata,
                    playlist_metadata,
                )

            if media_metadata["type"] in UPLOADED_VIDEO_MEDIA_TYPE:
                if not self.uploaded_video_downloader:
                    raise UnsupportedMediaType(media_metadata["type"])

                download_item = await self.uploaded_video_downloader.get_download_item(
                    media_metadata,
                )
        except Exception as e:
            download_item = DownloadItem(
                media_metadata=media_metadata,
                playlist_metadata=playlist_metadata,
                error=e,
            )

        return download_item

    async def get_collection_download_items(
        self,
        collection_metadata: dict,
    ) -> list[DownloadItem]:
        tracks_metadata = collection_metadata["relationships"]["tracks"]["data"]
        async for extended_data in self.interface.apple_music_api.extend_api_data(
            collection_metadata["relationships"]["tracks"],
        ):
            tracks_metadata.extend(extended_data["data"])

        tasks = [
            self.get_single_download_item(
                media_metadata,
                (
                    collection_metadata
                    if collection_metadata["type"] in PLAYLIST_MEDIA_TYPE
                    else None
                ),
            )
            for media_metadata in tracks_metadata
        ]

        download_items = await safe_gather(*tasks)
        return download_items

    async def get_artist_download_items(
        self,
        artist_metadata: dict,
    ) -> list[DownloadItem]:
        if not self.artist_auto_select:
            available_choices = []
            for artist_auto_select_option in list(ArtistAutoSelect):
                relation_key, type_key = artist_auto_select_option.path_key
                available_choices.append(
                    Choice(
                        name=str(artist_auto_select_option),
                        value=(artist_auto_select_option,),
                    ),
                )

            (artist_auto_select,) = await inquirer.select(
                message=f'Select which type to download for artist "{artist_metadata["attributes"]["name"]}":',
                choices=available_choices,
                validate=lambda result: artist_metadata.get(result[0].path_key[0], {})
                .get(result[0].path_key[1], {})
                .get("data"),
            ).execute_async()
        else:
            artist_auto_select = self.artist_auto_select

        relation_key, type_key = artist_auto_select.path_key
        artist_metadata[relation_key][type_key]["data"].extend(
            [
                extended_data
                async for extended_data in self.interface.apple_music_api.extend_api_data(
                    artist_metadata[relation_key][type_key],
                )
            ]
        )

        selected_items = artist_metadata[relation_key][type_key]["data"]
        select_all = self.artist_auto_select is not None

        if artist_auto_select in {
            ArtistAutoSelect.MAIN_ALBUMS,
            ArtistAutoSelect.COMPILATION_ALBUMS,
            ArtistAutoSelect.LIVE_ALBUMS,
            ArtistAutoSelect.SINGLES_EPS,
            ArtistAutoSelect.ALL_ALBUMS,
        }:
            return await self.get_artist_albums_download_items(
                selected_items,
                select_all,
            )
        elif artist_auto_select == ArtistAutoSelect.TOP_SONGS:
            return await self.get_artist_songs_download_items(
                selected_items,
                select_all,
            )
        elif artist_auto_select == ArtistAutoSelect.MUSIC_VIDEOS:
            return await self.get_artist_music_videos_download_items(
                selected_items,
                select_all,
            )

    async def get_artist_albums_download_items(
        self,
        albums_metadata: list[dict],
        select_all: bool = False,
    ) -> list[DownloadItem]:
        if not select_all:
            choices = [
                Choice(
                    name=" | ".join(
                        [
                            f'{album["attributes"]["trackCount"]:03d}',
                            f'{album["attributes"]["releaseDate"]:<10}',
                            f'{album["attributes"].get("contentRating", "None").title():<8}',
                            f'{album["attributes"]["name"]}',
                        ]
                    ),
                    value=album,
                )
                for album in albums_metadata
                if album.get("attributes")
            ]
            selected = await inquirer.select(
                message="Select which albums to download: (Track Count | Release Date | Rating | Title)",
                choices=choices,
                multiselect=True,
            ).execute_async()
        else:
            selected = albums_metadata

        download_items = []

        album_tasks = [
            self.interface.apple_music_api.get_album(album_metadata["id"])
            for album_metadata in selected
        ]
        album_responses = await safe_gather(*album_tasks)

        track_tasks = [
            self.get_collection_download_items(album_response["data"][0])
            for album_response in album_responses
        ]
        track_results = await safe_gather(*track_tasks)

        for track_result in track_results:
            download_items.extend(track_result)

        return download_items

    async def get_artist_music_videos_download_items(
        self,
        music_videos_metadata: list[dict],
        select_all: bool = False,
    ) -> list[DownloadItem]:
        if not select_all:
            choices = [
                Choice(
                    name=" | ".join(
                        [
                            self.millis_to_min_sec(
                                music_video["attributes"]["durationInMillis"]
                            ),
                            f'{music_video["attributes"].get("contentRating", "None").title():<8}',
                            music_video["attributes"]["name"],
                        ],
                    ),
                    value=music_video,
                )
                for music_video in music_videos_metadata
                if music_video.get("attributes")
            ]
            selected = await inquirer.select(
                message="Select which music videos to download: (Duration | Rating | Title)",
                choices=choices,
                multiselect=True,
            ).execute_async()
        else:
            selected = music_videos_metadata

        music_video_tasks = [
            self.get_single_download_item(
                music_video_metadata,
            )
            for music_video_metadata in selected
        ]
        download_items = await safe_gather(*music_video_tasks)

        return download_items

    async def get_artist_songs_download_items(
        self,
        songs_metadata: list[dict],
        select_all: bool = False,
    ) -> list[DownloadItem]:
        if not select_all:
            choices = [
                Choice(
                    name=" | ".join(
                        [
                            self.millis_to_min_sec(
                                song["attributes"]["durationInMillis"]
                            ),
                            f'{song["attributes"].get("contentRating", "None").title():<8}',
                            song["attributes"]["name"],
                        ],
                    ),
                    value=song,
                )
                for song in songs_metadata
                if song.get("attributes")
            ]
            selected = await inquirer.select(
                message="Select which songs to download: (Duration | Rating | Title)",
                choices=choices,
                multiselect=True,
            ).execute_async()
        else:
            selected = songs_metadata

        song_tasks = [
            self.get_single_download_item(
                song_metadata,
            )
            for song_metadata in selected
        ]
        download_items = await safe_gather(*song_tasks)

        return download_items

    def millis_to_min_sec(self, millis) -> str:
        minutes, seconds = divmod(millis // 1000, 60)
        return f"{minutes:02}:{seconds:02}"

    def get_url_info(self, url: str) -> UrlInfo | None:
        match = VALID_URL_PATTERN.match(url)
        if not match:
            return None

        return UrlInfo(
            **match.groupdict(),
        )

    async def get_download_queue(
        self,
        url_info: UrlInfo,
    ) -> list[DownloadItem] | None:
        return await self._get_download_queue(
            "song" if url_info.sub_id else url_info.type or url_info.library_type,
            url_info.sub_id or url_info.id or url_info.library_id,
            url_info.library_id is not None,
        )

    async def _get_download_queue(
        self,
        url_type: str,
        id: str,
        is_library: bool,
    ) -> list[DownloadItem] | None:
        download_items = []

        if url_type in ARTIST_MEDIA_TYPE:
            try:
                artist_response = await self.interface.apple_music_api.get_artist(
                    id,
                )
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if artist_response is None:
                return None

            download_items = await self.get_artist_download_items(
                artist_response["data"][0],
            )

        if url_type in SONG_MEDIA_TYPE:
            try:
                song_respose = await self.interface.apple_music_api.get_song(id)
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if song_respose is None:
                return None

            download_items.append(
                await self.get_single_download_item(song_respose["data"][0])
            )

        if url_type in ALBUM_MEDIA_TYPE:
            try:
                if is_library:
                    album_response = (
                        await self.interface.apple_music_api.get_library_album(id)
                    )
                else:
                    album_response = await self.interface.apple_music_api.get_album(id)
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if album_response is None:
                return None

            download_items = await self.get_collection_download_items(
                album_response["data"][0],
            )

        if url_type in PLAYLIST_MEDIA_TYPE:
            try:
                if is_library:
                    playlist_response = (
                        await self.interface.apple_music_api.get_library_playlist(id)
                    )
                else:
                    playlist_response = (
                        await self.interface.apple_music_api.get_playlist(id)
                    )
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if playlist_response is None:
                return None

            download_items = await self.get_collection_download_items(
                playlist_response["data"][0],
            )

        if url_type in MUSIC_VIDEO_MEDIA_TYPE:
            try:
                music_video_response = (
                    await self.interface.apple_music_api.get_music_video(id)
                )
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if music_video_response is None:
                return None

            download_items.append(
                await self.get_single_download_item(music_video_response["data"][0])
            )

        if url_type in UPLOADED_VIDEO_MEDIA_TYPE:
            try:
                uploaded_video = (
                    await self.interface.apple_music_api.get_uploaded_video(id)
                )
            except ApiError as e:
                if e.status_code == 404:
                    return None
                raise e

            if uploaded_video is None:
                return None

            download_items.append(
                await self.get_single_download_item(uploaded_video["data"][0])
            )

        return download_items

    async def download(
        self,
        download_item: DownloadItem,
    ) -> DownloadItem:
        try:
            if download_item.flat_filter_result:
                download_item = await self.get_single_download_item_no_filter(
                    download_item.media_metadata,
                    download_item.playlist_metadata,
                )

            if download_item.error:
                raise download_item.error

            await self._initial_processing(download_item)
            await self._download(download_item)
            await self._final_processing(download_item)

            return download_item
        finally:
            if isinstance(download_item, DownloadItem) and not self.skip_processing:
                self.base_downloader.cleanup_temp(download_item.random_uuid)

    async def _download(
        self,
        download_item: DownloadItem,
    ) -> None:
        if (
            self.song_downloader.synced_lyrics_only
            and download_item.media_metadata["type"] not in SONG_MEDIA_TYPE
        ):
            raise SyncedLyricsOnly()

        if self.song_downloader.synced_lyrics_only:
            return

        if (
            Path(download_item.final_path).exists()
            and not self.base_downloader.overwrite
        ):
            raise MediaFileExists(download_item.final_path)

        if (
            self.base_downloader.download_mode == DownloadMode.NM3U8DLRE
            and not self.base_downloader.full_nm3u8dlre_path
        ):
            raise ExecutableNotFound("N_m3u8DL-RE")

        if download_item.media_metadata["type"] in MUSIC_VIDEO_MEDIA_TYPE:
            if (
                self.music_video_downloader.remux_mode == RemuxMode.FFMPEG
                and not self.base_downloader.full_ffmpeg_path
            ):
                raise ExecutableNotFound("ffmpeg")

            if (
                self.music_video_downloader.remux_mode == RemuxMode.MP4BOX
                and not self.base_downloader.full_mp4box_path
            ):
                raise ExecutableNotFound("MP4Box")

            if (
                download_item.media_metadata["type"] in MUSIC_VIDEO_MEDIA_TYPE
                or self.music_video_downloader.remux_mode == RemuxMode.MP4BOX
            ) and not self.base_downloader.full_mp4decrypt_path:
                raise ExecutableNotFound("mp4decrypt")

        if (
            not download_item.stream_info
            or not download_item.stream_info.audio_track
            or not download_item.stream_info.audio_track.stream_url
            or (
                (
                    not download_item.decryption_key
                    or not download_item.decryption_key.audio_track
                    or not download_item.decryption_key.audio_track.key
                )
                and not self.base_downloader.use_wrapper
            )
        ):
            raise FormatNotAvailable(download_item.media_metadata["id"])

        if download_item.media_metadata["type"] in SONG_MEDIA_TYPE:
            await self.song_downloader.download(download_item)

        if download_item.media_metadata["type"] in MUSIC_VIDEO_MEDIA_TYPE:
            await self.music_video_downloader.download(download_item)

        if download_item.media_metadata["type"] in UPLOADED_VIDEO_MEDIA_TYPE:
            await self.uploaded_video_downloader.download(download_item)

    async def _initial_processing(
        self,
        download_item: DownloadItem,
    ) -> None:
        if self.skip_processing:
            return

        if download_item.cover_path and self.base_downloader.save_cover:
            cover_bytes = await self.interface.get_cover_bytes(download_item.cover_url)
            if cover_bytes and (
                self.base_downloader.overwrite
                or not Path(download_item.cover_path).exists()
            ):
                self.base_downloader.write_cover_image(
                    cover_bytes,
                    download_item.cover_path,
                )

        if (
            download_item.lyrics
            and download_item.lyrics.synced
            and not self.song_downloader.no_synced_lyrics
            and (
                self.base_downloader.overwrite
                or not Path(download_item.synced_lyrics_path).exists()
            )
        ):
            self.song_downloader.write_synced_lyrics(
                download_item.lyrics.synced,
                download_item.synced_lyrics_path,
            )

        if download_item.playlist_tags and self.base_downloader.save_playlist:
            self.base_downloader.update_playlist_file(
                download_item.playlist_file_path,
                download_item.final_path,
                download_item.playlist_tags.playlist_track,
            )

    async def _final_processing(
        self,
        download_item: DownloadItem,
    ) -> None:
        if self.skip_processing:
            return

        if download_item.staged_path and Path(download_item.staged_path).exists():
            self.base_downloader.move_to_final_path(
                download_item.staged_path,
                download_item.final_path,
            )



"""
This is a modified version of https://github.com/sn0wst0rm/st0rmMusicPlayer/blob/main/scripts/amdecrypt.py
All the modifications made here were AI generated
"""

import asyncio
import io
import logging
import struct
import time
from dataclasses import dataclass, field
from typing import BinaryIO, List, Optional

from Crypto.Cipher import AES

logger = logging.getLogger(__name__)

# Default decryption key for songs without per-sample keys (legacy AAC)
DEFAULT_SONG_DECRYPTION_KEY = b"2\xb8\xad\xe1v\x9e&\xb1\xff\xb8\x98cRy?\xc6"

# Pre-fetch key used for first sample description
PREFETCH_KEY = "skd://itunes.apple.com/P000000000/s1/e1"

# Default wrapper address
DEFAULT_WRAPPER_IP = "127.0.0.1:10020"


@dataclass
class SampleInfo:
    """Information about a single audio sample."""

    data: bytes
    duration: int
    desc_index: int
    iv: bytes = b""  # Per-sample IV from senc (empty if constant IV)
    subsamples: List[tuple] = field(
        default_factory=list
    )  # [(clear_bytes, encrypted_bytes), ...]


@dataclass
class EncryptionInfo:
    """Encryption scheme info extracted from sinf/schm + sinf/schi/tenc."""

    scheme_type: str = "cbcs"  # 'cenc' or 'cbcs'
    per_sample_iv_size: int = 0  # 0, 8, or 16
    constant_iv: bytes = b""  # Constant IV (when per_sample_iv_size == 0)
    kid: bytes = b""  # Default Key ID (16 bytes)


@dataclass
class SongInfo:
    """Extracted song information from MP4 file."""

    samples: List[SampleInfo] = field(default_factory=list)
    moov_data: bytes = b""
    ftyp_data: bytes = b""
    encryption_info: Optional[EncryptionInfo] = None


def read_box_header(f: BinaryIO) -> tuple[int, str, int]:
    """Read MP4 box header, return (size, type, header_size)."""
    header = f.read(8)
    if len(header) < 8:
        return 0, "", 0

    size = struct.unpack(">I", header[:4])[0]
    box_type = header[4:8].decode("ascii", errors="replace")
    header_size = 8

    if size == 1:  # Extended size
        ext_size = f.read(8)
        size = struct.unpack(">Q", ext_size)[0]
        header_size = 16
    elif size == 0:  # Box extends to end of file
        pos = f.tell()
        f.seek(0, 2)  # Seek to end
        size = f.tell() - pos + header_size
        f.seek(pos)

    return size, box_type, header_size


def find_box(data: bytes, box_path: List[str]) -> Optional[bytes]:
    """Find a box in MP4 data by path (e.g., ['moov', 'trak', 'mdia'])."""
    f = io.BytesIO(data)

    for target_type in box_path:
        found = False
        while True:
            pos = f.tell()
            size, box_type, header_size = read_box_header(f)
            if size == 0:
                break

            if box_type == target_type:
                f.seek(pos + header_size)  # Skip past header
                found = True
                break
            else:
                f.seek(pos + size)  # Skip this box

        if not found:
            return None

    # Return remaining data from current position
    return f.read()


def extract_song(input_path: str) -> SongInfo:
    """
    Extract song samples and metadata from encrypted MP4 file.

    This parses the MP4 structure to extract:
    - ftyp and moov boxes (for reassembly)
    - Individual audio samples from mdat boxes
    - Sample durations and description indices from moof boxes
    """
    with open(input_path, "rb") as f:
        raw_data = f.read()

    song_info = SongInfo()

    # First pass: collect all top-level boxes
    boxes = []
    offset = 0
    while offset < len(raw_data) - 8:
        size = struct.unpack(">I", raw_data[offset : offset + 4])[0]
        box_type = raw_data[offset + 4 : offset + 8].decode("ascii", errors="replace")

        header_size = 8
        if size == 0:
            break
        if size == 1:
            # Extended size
            if offset + 16 > len(raw_data):
                break
            size = struct.unpack(">Q", raw_data[offset + 8 : offset + 16])[0]
            header_size = 16

        boxes.append(
            {
                "offset": offset,
                "size": size,
                "type": box_type,
                "header_size": header_size,
                "data": raw_data[offset : offset + size],
            }
        )
        offset += size

    logger.debug(f"Found {len(boxes)} top-level boxes")

    # Extract ftyp and moov
    for box in boxes:
        if box["type"] == "ftyp":
            song_info.ftyp_data = box["data"]
        elif box["type"] == "moov":
            song_info.moov_data = box["data"]

    # Get default sample info from trex (inside moov)
    default_sample_duration = 1024
    default_sample_size = 0

    # Determine which track is the audio track
    audio_track_id = (
        _extract_audio_track_id(song_info.moov_data) if song_info.moov_data else 1
    )
    logger.debug(f"Audio track ID: {audio_track_id}")

    # Extract encryption scheme info from moov (sinf/schm + sinf/schi/tenc)
    if song_info.moov_data:
        song_info.encryption_info = _extract_encryption_info(song_info.moov_data)

    # Parse moof/mdat pairs
    moof_box = None
    for box in boxes:
        if box["type"] == "moof":
            moof_box = box
        elif box["type"] == "mdat" and moof_box is not None:
            # Parse this moof/mdat pair
            moof_data = moof_box["data"]
            mdat_data = box["data"][box["header_size"] :]  # Skip mdat header

            # Parse moof for tfhd (sample description index, defaults) and trun (entries)
            _iv_size = (
                song_info.encryption_info.per_sample_iv_size
                if song_info.encryption_info
                else 0
            )
            samples_from_pair = _parse_moof_mdat(
                moof_data,
                mdat_data,
                default_sample_duration,
                default_sample_size,
                audio_track_id=audio_track_id,
                moof_offset=moof_box["offset"],
                mdat_data_offset=box["offset"] + box["header_size"],
                per_sample_iv_size=_iv_size,
            )
            song_info.samples.extend(samples_from_pair)
            moof_box = None

    logger.debug(f"Extracted {len(song_info.samples)} samples from {input_path}")
    return song_info


def _parse_moof_mdat(
    moof_data: bytes,
    mdat_data: bytes,
    default_sample_duration: int,
    default_sample_size: int,
    audio_track_id: int = 1,
    moof_offset: int = 0,
    mdat_data_offset: int = 0,
    per_sample_iv_size: int = 0,
) -> List[SampleInfo]:
    """Parse a moof box and extract samples from corresponding mdat.

    Handles multi-track fragmented MP4s by only extracting samples from
    the traf matching the audio track ID.

    Args:
        audio_track_id: Track ID of the audio track to extract.
        moof_offset: Absolute file offset of the moof box.
        mdat_data_offset: Absolute file offset of the mdat content (after header).
        per_sample_iv_size: IV size per sample from tenc (0, 8, or 16).
    """
    samples = []

    # Simple box parsing inside moof
    offset = 8  # Skip moof header
    while offset < len(moof_data) - 8:
        size = struct.unpack(">I", moof_data[offset : offset + 4])[0]
        box_type = moof_data[offset + 4 : offset + 8].decode("ascii", errors="replace")

        if size == 0 or offset + size > len(moof_data):
            break

        if box_type == "traf":
            # Parse inside traf with per-traf state
            tfhd_info = {
                "track_id": 0,
                "desc_index": 0,
                "default_duration": default_sample_duration,
                "default_size": default_sample_size,
                "flags": 0,
                "base_data_offset": None,
            }
            trun_entries = []
            first_trun_data_offset = None
            senc_entries = []  # Per-sample encryption info from senc box

            traf_offset = offset + 8
            traf_end = offset + size
            while traf_offset < traf_end - 8:
                inner_size = struct.unpack(
                    ">I", moof_data[traf_offset : traf_offset + 4]
                )[0]
                inner_type = moof_data[traf_offset + 4 : traf_offset + 8].decode(
                    "ascii", errors="replace"
                )

                if inner_size == 0:
                    break

                if inner_type == "tfhd":
                    _parse_tfhd(
                        moof_data[traf_offset + 8 : traf_offset + inner_size], tfhd_info
                    )
                elif inner_type == "trun":
                    entries, data_off = _parse_trun(
                        moof_data[traf_offset + 8 : traf_offset + inner_size], tfhd_info
                    )
                    if first_trun_data_offset is None:
                        first_trun_data_offset = data_off
                    trun_entries.extend(entries)
                elif inner_type == "senc":
                    senc_entries = _parse_senc(
                        moof_data[traf_offset + 8 : traf_offset + inner_size],
                        per_sample_iv_size,
                    )

                traf_offset += inner_size

            # Only process this traf if it matches the audio track
            if tfhd_info["track_id"] != audio_track_id:
                offset += size
                continue

            # Compute starting offset in mdat_data
            base = tfhd_info.get("base_data_offset")
            if base is None:
                base = moof_offset  # Default: first byte of containing moof

            if first_trun_data_offset is not None:
                mdat_idx = (base + first_trun_data_offset) - mdat_data_offset
            else:
                mdat_idx = 0

            mdat_read_offset = max(0, mdat_idx)
            desc_index = tfhd_info["desc_index"]
            if desc_index > 0:
                desc_index -= 1  # Convert to 0-indexed

            for i, entry in enumerate(trun_entries):
                sample_size = entry.get("size", tfhd_info["default_size"])
                sample_duration = entry.get("duration", tfhd_info["default_duration"])

                if sample_size > 0 and mdat_read_offset + sample_size <= len(mdat_data):
                    # Attach per-sample encryption info from senc if available
                    sample_iv = b""
                    sample_subsamples = []
                    if i < len(senc_entries):
                        sample_iv = senc_entries[i]["iv"]
                        sample_subsamples = senc_entries[i]["subsamples"]

                    sample = SampleInfo(
                        data=mdat_data[
                            mdat_read_offset : mdat_read_offset + sample_size
                        ],
                        duration=sample_duration,
                        desc_index=desc_index,
                        iv=sample_iv,
                        subsamples=sample_subsamples,
                    )
                    samples.append(sample)
                    mdat_read_offset += sample_size

        offset += size

    return samples


def _parse_tfhd(data: bytes, tfhd_info: dict):
    """Parse track fragment header box (FullBox: version + flags + content)."""
    if len(data) < 8:  # version(1) + flags(3) + track_id(4)
        return

    # FullBox: version(1) + flags(3)
    version = data[0]
    flags = struct.unpack(">I", b"\x00" + data[1:4])[0]
    tfhd_info["flags"] = flags

    # After version+flags is track_id(4)
    tfhd_info["track_id"] = struct.unpack(">I", data[4:8])[0]
    offset = 4 + 4  # version+flags + track_id

    if flags & 0x01 and offset + 8 <= len(data):  # base_data_offset
        tfhd_info["base_data_offset"] = struct.unpack(">Q", data[offset : offset + 8])[
            0
        ]
        offset += 8
    if flags & 0x02 and offset + 4 <= len(data):  # sample_description_index
        tfhd_info["desc_index"] = struct.unpack(">I", data[offset : offset + 4])[0]
        offset += 4
    if flags & 0x08 and offset + 4 <= len(data):  # default_sample_duration
        tfhd_info["default_duration"] = struct.unpack(">I", data[offset : offset + 4])[
            0
        ]
        offset += 4
    if flags & 0x10 and offset + 4 <= len(data):  # default_sample_size
        tfhd_info["default_size"] = struct.unpack(">I", data[offset : offset + 4])[0]


def _parse_trun(data: bytes, tfhd_info: dict) -> tuple[List[dict], Optional[int]]:
    """Parse track run box to get sample entries and data_offset.

    Returns:
        Tuple of (entries, data_offset). data_offset is the signed offset from
        base_data_offset to the first sample's data, or None if not present.
    """
    entries = []
    data_offset_value = None
    if len(data) < 8:  # version(1) + flags(3) + sample_count(4)
        return entries, data_offset_value

    # FullBox: version(1) + flags(3)
    version = data[0]
    flags = struct.unpack(">I", b"\x00" + data[1:4])[0]
    sample_count = struct.unpack(">I", data[4:8])[0]

    # Start reading entries after header fields
    offset = 8  # version+flags(4) + sample_count(4)
    if flags & 0x01:  # data_offset present
        data_offset_value = struct.unpack(">i", data[offset : offset + 4])[0]
        offset += 4
    if flags & 0x04:  # first_sample_flags present
        offset += 4

    for _ in range(sample_count):
        entry = {}
        if flags & 0x100 and offset + 4 <= len(data):  # sample_duration
            entry["duration"] = struct.unpack(">I", data[offset : offset + 4])[0]
            offset += 4
        if flags & 0x200 and offset + 4 <= len(data):  # sample_size
            entry["size"] = struct.unpack(">I", data[offset : offset + 4])[0]
            offset += 4
        if flags & 0x400:  # sample_flags
            offset += 4
        if flags & 0x800:  # sample_composition_time_offset
            offset += 4
        entries.append(entry)

    return entries, data_offset_value


def _parse_senc(data: bytes, per_sample_iv_size: int) -> List[dict]:
    """Parse Sample Encryption Box (senc) content (after box header).

    Returns a list of dicts, one per sample:
        {"iv": bytes, "subsamples": [(clear_bytes, encrypted_bytes), ...]}

    The data starts after the 8-byte box header (size+type) but includes
    the FullBox header (version 1 byte + flags 3 bytes).

    per_sample_iv_size can be 0 (cbcs with constant IV from tenc) — in that case
    there are 0 IV bytes per sample but subsample info may still be present.
    """
    if len(data) < 8:
        return []

    version = data[0]
    flags = struct.unpack(">I", b"\x00" + data[1:4])[0]
    sample_count = struct.unpack(">I", data[4:8])[0]

    entries: List[dict] = []
    offset = 8
    for _ in range(sample_count):
        # Read per-sample IV (0 bytes when per_sample_iv_size == 0)
        iv = b""
        if per_sample_iv_size > 0:
            if offset + per_sample_iv_size > len(data):
                break
            iv = data[offset : offset + per_sample_iv_size]
            offset += per_sample_iv_size

        subsamples = []
        if flags & 0x02:
            # Subsample encryption info present
            if offset + 2 > len(data):
                break
            subsample_count = struct.unpack(">H", data[offset : offset + 2])[0]
            offset += 2
            for _ in range(subsample_count):
                if offset + 6 > len(data):
                    break
                clear_bytes = struct.unpack(">H", data[offset : offset + 2])[0]
                encrypted_bytes = struct.unpack(">I", data[offset + 2 : offset + 6])[0]
                subsamples.append((clear_bytes, encrypted_bytes))
                offset += 6

        entries.append({"iv": iv, "subsamples": subsamples})

    return entries


async def decrypt_samples(
    wrapper_ip: str,
    track_id: str,
    fairplay_key: str,
    samples: List[SampleInfo],
    progress_callback=None,
) -> bytes:
    """
    Send samples to wrapper for CBCS decryption and return decrypted data.

    CBCS full subsample encryption (used by ALAC):
    - Only bytes aligned to 16 are encrypted
    - Remaining bytes (len % 16) are clear and kept as-is

    Protocol:
    - For each new key: [1B id_len][id][1B key_len][key]
    - For each sample: [4B LE truncated_size][truncated_data] -> read back [decrypted_data]
    - Key switch: [0,0,0,0]
    - Close: [0,0,0,0,0]

    Args:
        progress_callback: Optional callback(current_sample, total_samples, bytes_processed) for progress tracking
    """
    host, port = wrapper_ip.split(":")
    port = int(port)

    reader, writer = await asyncio.open_connection(host, port)

    try:
        decrypted_data = bytearray()
        last_desc_index = 255

        keys = [PREFETCH_KEY, fairplay_key]
        total_samples = len(samples)
        bytes_processed = 0
        start_time = time.time()
        last_progress_time = start_time

        for i, sample in enumerate(samples):
            # Check if we need to switch keys
            if last_desc_index != sample.desc_index:
                if last_desc_index != 255:
                    # Send key switch signal
                    writer.write(struct.pack("<I", 0))
                    await writer.drain()

                # Send new key info
                key_uri = keys[min(sample.desc_index, len(keys) - 1)]

                if key_uri == PREFETCH_KEY:
                    id_bytes = b"0"
                else:
                    id_bytes = track_id.encode("utf-8")
                writer.write(struct.pack("B", len(id_bytes)))
                writer.write(id_bytes)

                key_bytes = key_uri.encode("utf-8")
                writer.write(struct.pack("B", len(key_bytes)))
                writer.write(key_bytes)
                await writer.drain()

                last_desc_index = sample.desc_index

            # CBCS full subsample decryption: truncate to 16-byte boundary
            sample_len = len(sample.data)
            truncated_len = sample_len & ~0xF

            if truncated_len > 0:
                # Send size and data
                writer.write(struct.pack("<I", truncated_len))
                writer.write(sample.data[:truncated_len])
                await writer.drain()

                # Read decrypted data
                decrypted_sample = await reader.readexactly(truncated_len)
                if len(decrypted_sample) != truncated_len:
                    raise IOError(
                        f"Short read: got {len(decrypted_sample)}, expected {truncated_len}"
                    )
                decrypted_data.extend(decrypted_sample)
                bytes_processed += truncated_len

            # Append clear bytes
            if truncated_len < sample_len:
                decrypted_data.extend(sample.data[truncated_len:])
                bytes_processed += sample_len - truncated_len

            # Call progress callback every 50 samples or 0.5s
            now = time.time()
            if progress_callback and (
                i % 50 == 0 or now - last_progress_time > 0.5 or i == total_samples - 1
            ):
                elapsed = now - start_time
                speed = bytes_processed / elapsed if elapsed > 0 else 0
                progress_callback(i + 1, total_samples, bytes_processed, speed)
                last_progress_time = now

        # Send close signal
        writer.write(bytes([0, 0, 0, 0, 0]))
        await writer.drain()

        logger.debug(f"Decrypted {len(samples)} samples ({len(decrypted_data)} bytes)")
        return bytes(decrypted_data)

    finally:
        writer.close()
        await writer.wait_closed()


def write_decrypted_m4a(
    output_path: str,
    song_info: SongInfo,
    decrypted_data: bytes,
    original_path: str = None,
) -> None:
    """
    Write decrypted MP4 file as non-fragmented MP4.

    Creates a new MP4 from scratch with:
    - ftyp box (M4A compatible)
    - moov box with proper sample tables (stts, stsc, stsz, stco)
    - Single mdat box with all decrypted samples

    This matches the output format of Go's amdecrypt which is required
    for ALAC playback.
    """
    # Extract original boxes for faithful reproduction
    # Note: _extract_stsd_content automatically cleans encryption metadata
    stsd_content = None
    orig_mvhd = None
    orig_tkhd = None
    orig_mdhd = None
    orig_hdlr = None
    orig_smhd = None
    orig_dinf = None
    timescale = 44100  # Default

    if original_path:
        with open(original_path, "rb") as f:
            orig_data = f.read()
    elif song_info.moov_data:
        orig_data = song_info.ftyp_data + song_info.moov_data
    else:
        orig_data = None

    if orig_data:
        stsd_content = _extract_stsd_content(orig_data)
        timescale = _extract_timescale(orig_data)

        # Find moov box and extract child boxes
        moov_idx = orig_data.find(b"moov")
        if moov_idx >= 4:
            moov_size = struct.unpack(">I", orig_data[moov_idx - 4 : moov_idx])[0]
            moov_data = orig_data[moov_idx - 4 : moov_idx - 4 + moov_size]

            orig_mvhd = _find_child_box(moov_data, b"mvhd")

            audio_trak = _find_audio_trak(moov_data)
            if audio_trak:
                orig_tkhd = _find_child_box(audio_trak, b"tkhd")
                mdia = _find_child_box(audio_trak, b"mdia")
                if mdia:
                    orig_mdhd = _find_child_box(mdia, b"mdhd")
                    orig_hdlr = _find_child_box(mdia, b"hdlr")
                    minf = _find_child_box(mdia, b"minf")
                    if minf:
                        orig_smhd = _find_child_box(minf, b"smhd")
                        orig_dinf = _find_child_box(minf, b"dinf")

    with open(output_path, "wb") as f:
        # Write ftyp
        _write_ftyp(f)

        # Calculate total duration
        total_duration = sum(s.duration for s in song_info.samples)

        # Write moov with sample tables
        _write_moov(
            f,
            song_info.samples,
            total_duration,
            timescale,
            stsd_content,
            decrypted_data,
            orig_mvhd=orig_mvhd,
            orig_tkhd=orig_tkhd,
            orig_mdhd=orig_mdhd,
            orig_hdlr=orig_hdlr,
            orig_smhd=orig_smhd,
            orig_dinf=orig_dinf,
        )

        # Write mdat
        _write_mdat(f, decrypted_data)

    logger.debug(f"Wrote decrypted file to {output_path}")


def _write_box(f, box_type: bytes, content: bytes):
    """Write a simple MP4 box."""
    size = len(content) + 8
    f.write(struct.pack(">I", size))
    f.write(box_type)
    f.write(content)


def _write_ftyp(f):
    """Write ftyp box for M4A."""
    content = b"M4A " + struct.pack(">I", 0)  # major brand + minor version
    content += b"M4A mp42isom\x00\x00\x00\x00"  # compatible brands
    _write_box(f, b"ftyp", content)


def _write_fullbox(f, box_type: bytes, version: int, flags: int, content: bytes):
    """Write a FullBox (with version and flags)."""
    size = len(content) + 12
    f.write(struct.pack(">I", size))
    f.write(box_type)
    f.write(struct.pack("B", version))
    f.write(struct.pack(">I", flags)[1:])  # 3 bytes for flags
    f.write(content)


def _write_moov(
    f,
    samples: List[SampleInfo],
    total_duration: int,
    timescale: int,
    stsd_content: bytes,
    decrypted_data: bytes,
    orig_mvhd: Optional[bytes] = None,
    orig_tkhd: Optional[bytes] = None,
    orig_mdhd: Optional[bytes] = None,
    orig_hdlr: Optional[bytes] = None,
    orig_smhd: Optional[bytes] = None,
    orig_dinf: Optional[bytes] = None,
):
    """Write moov box with sample tables.

    When original box data is available, it is copied verbatim (with duration
    fields patched) to faithfully reproduce the source file's metadata.
    This matches the Go amdecrypt behavior of copying boxes from the original.
    """
    moov_start = f.tell()
    f.write(b"\x00" * 8)  # moov header placeholder

    # mvhd (movie header)
    if orig_mvhd:
        f.write(_patch_mvhd_duration(orig_mvhd, total_duration))
    else:
        mvhd_content = struct.pack(">II", 0, 0)  # creation, modification
        mvhd_content += struct.pack(">I", timescale)
        mvhd_content += struct.pack(">I", total_duration)
        mvhd_content += struct.pack(">I", 0x00010000)  # rate (1.0)
        mvhd_content += struct.pack(">H", 0x0100)  # volume (1.0)
        mvhd_content += b"\x00" * 10  # reserved
        mvhd_content += struct.pack(
            ">9I", 0x00010000, 0, 0, 0, 0x00010000, 0, 0, 0, 0x40000000
        )  # matrix
        mvhd_content += b"\x00" * 24  # pre_defined
        mvhd_content += struct.pack(">I", 2)  # next_track_id
        _write_fullbox(f, b"mvhd", 0, 0, mvhd_content)

    # trak (track)
    trak_start = f.tell()
    f.write(b"\x00" * 8)  # trak header placeholder

    # tkhd (track header)
    if orig_tkhd:
        f.write(_patch_tkhd_duration(orig_tkhd, total_duration))
    else:
        tkhd_content = struct.pack(">II", 0, 0)  # creation, modification
        tkhd_content += struct.pack(">I", 1)  # track_id
        tkhd_content += struct.pack(">I", 0)  # reserved
        tkhd_content += struct.pack(">I", total_duration)
        tkhd_content += b"\x00" * 8  # reserved
        tkhd_content += struct.pack(">HH", 0, 0)  # layer, alternate_group
        tkhd_content += struct.pack(">H", 0x0100)  # volume
        tkhd_content += struct.pack(">H", 0)  # reserved
        tkhd_content += struct.pack(
            ">9I", 0x00010000, 0, 0, 0, 0x00010000, 0, 0, 0, 0x40000000
        )  # matrix
        tkhd_content += struct.pack(">II", 0, 0)  # width, height
        _write_fullbox(f, b"tkhd", 0, 7, tkhd_content)

    # mdia (media)
    mdia_start = f.tell()
    f.write(b"\x00" * 8)

    # mdhd (media header) - preserves original language code
    if orig_mdhd:
        f.write(_patch_mdhd_duration(orig_mdhd, total_duration))
    else:
        mdhd_content = struct.pack(">II", 0, 0)  # creation, modification
        mdhd_content += struct.pack(">I", timescale)
        mdhd_content += struct.pack(">I", total_duration)
        mdhd_content += struct.pack(">H", 0x55C4)  # language (und)
        mdhd_content += struct.pack(">H", 0)  # quality
        _write_fullbox(f, b"mdhd", 0, 0, mdhd_content)

    # hdlr (handler) - preserves original handler name (e.g. "Core Media Audio")
    if orig_hdlr:
        f.write(orig_hdlr)
    else:
        hdlr_content = struct.pack(">I", 0)  # pre_defined
        hdlr_content += b"soun"  # handler_type
        hdlr_content += b"\x00" * 12  # reserved
        handler_name = b"SoundHandler"
        hdlr_content += struct.pack("B", len(handler_name)) + handler_name + b"\x00"
        _write_fullbox(f, b"hdlr", 0, 0, hdlr_content)

    # minf (media info)
    minf_start = f.tell()
    f.write(b"\x00" * 8)

    # smhd (sound media header)
    if orig_smhd:
        f.write(orig_smhd)
    else:
        smhd_content = struct.pack(">HH", 0, 0)  # balance, reserved
        _write_fullbox(f, b"smhd", 0, 0, smhd_content)

    # dinf + dref
    if orig_dinf:
        f.write(orig_dinf)
    else:
        dinf_start = f.tell()
        f.write(b"\x00" * 8)
        dref_content = struct.pack(">I", 1)  # entry_count
        dref_content += (
            struct.pack(">I", 12) + b"url " + struct.pack(">I", 1)
        )  # url entry (self-contained)
        _write_fullbox(f, b"dref", 0, 0, dref_content)
        _fixup_box_size(f, dinf_start, b"dinf")

    # stbl (sample table)
    stbl_start = f.tell()
    f.write(b"\x00" * 8)

    # stsd (sample description) - use content from original file
    _write_stsd(f, stsd_content)

    # stts (time-to-sample)
    _write_stts(f, samples)

    # stsc (sample-to-chunk) - all samples in one chunk
    stsc_content = struct.pack(">I", 1)  # entry_count
    stsc_content += struct.pack(
        ">III", 1, len(samples), 1
    )  # first_chunk, samples_per_chunk, sample_description_index
    _write_fullbox(f, b"stsc", 0, 0, stsc_content)

    # stsz (sample size)
    stsz_content = struct.pack(">I", 0)  # sample_size (0 = variable)
    stsz_content += struct.pack(">I", len(samples))  # sample_count
    for sample in samples:
        stsz_content += struct.pack(">I", len(sample.data))
    _write_fullbox(f, b"stsz", 0, 0, stsz_content)

    # stco (chunk offset) - will be fixed up later
    stco_pos = f.tell()
    stco_content = struct.pack(">I", 1)  # entry_count
    stco_content += struct.pack(">I", 0)  # chunk_offset (placeholder)
    _write_fullbox(f, b"stco", 0, 0, stco_content)

    _fixup_box_size(f, stbl_start, b"stbl")
    _fixup_box_size(f, minf_start, b"minf")
    _fixup_box_size(f, mdia_start, b"mdia")
    _fixup_box_size(f, trak_start, b"trak")

    # udta > meta > hdlr(mdir) + ilst - metadata container
    _write_udta(f)

    _fixup_box_size(f, moov_start, b"moov")

    # Fix up stco with correct mdat offset
    mdat_offset = f.tell() + 8  # +8 for mdat header
    f.seek(stco_pos + 16)  # +12 for box header + version/flags, +4 for entry_count
    f.write(struct.pack(">I", mdat_offset))
    f.seek(0, 2)  # Back to end


def _write_stsd(f, stsd_content: bytes):
    """Write sample description box using content from original file.
    This preserves the original codec info (ALAC, EC-3, AAC, etc.).
    """
    if stsd_content:
        # Write the full stsd box with its content from the source file
        size = len(stsd_content) + 8
        f.write(struct.pack(">I", size))
        f.write(b"stsd")
        f.write(stsd_content)
    else:
        # Fallback: write a basic ALAC stsd if no source content available
        _write_stsd_alac_fallback(f)


def _write_stsd_alac_fallback(f):
    """Write a default ALAC sample description box (fallback)."""
    stsd_start = f.tell()
    f.write(b"\x00" * 12)  # box header + version/flags placeholder

    f.write(struct.pack(">I", 1))  # entry_count

    # alac sample entry
    alac_start = f.tell()
    f.write(b"\x00" * 8)  # alac box header placeholder

    f.write(b"\x00" * 6)  # reserved
    f.write(struct.pack(">H", 1))  # data_reference_index
    f.write(b"\x00" * 8)  # reserved
    f.write(struct.pack(">H", 2))  # channel_count
    f.write(struct.pack(">H", 16))  # sample_size (bits)
    f.write(struct.pack(">H", 0))  # pre_defined
    f.write(struct.pack(">H", 0))  # reserved
    f.write(struct.pack(">I", 44100 << 16))  # sample_rate (16.16 fixed point)

    # alac magic cookie box
    # Default ALAC config for 44.1kHz stereo 24-bit
    default_config = bytes(
        [
            0x00,
            0x00,
            0x10,
            0x00,  # frame_length
            0x00,  # compatible_version
            0x18,  # bit_depth (24)
            0x28,
            0x28,
            0x0A,  # pb, mb, kb
            0x02,  # num_channels
            0x00,
            0x00,  # max_run
            0x00,
            0x00,
            0xFF,
            0xFF,  # max_frame_bytes
            0x00,
            0x0D,
            0x00,
            0x80,  # avg_bit_rate
            0x00,
            0x00,
            0xAC,
            0x44,  # sample_rate
        ]
    )
    _write_box(f, b"alac", default_config)

    _fixup_box_size(f, alac_start, b"alac")

    # Fix stsd size
    end_pos = f.tell()
    size = end_pos - stsd_start
    f.seek(stsd_start)
    f.write(struct.pack(">I", size))
    f.write(b"stsd")
    f.write(struct.pack(">I", 0))  # version + flags
    f.seek(end_pos)


def _write_stts(f, samples: List[SampleInfo]):
    """Write time-to-sample box (run-length encoded)."""
    # Run-length encode durations
    entries = []
    for sample in samples:
        if entries and entries[-1][1] == sample.duration:
            entries[-1] = (entries[-1][0] + 1, sample.duration)
        else:
            entries.append((1, sample.duration))

    content = struct.pack(">I", len(entries))
    for count, delta in entries:
        content += struct.pack(">II", count, delta)
    _write_fullbox(f, b"stts", 0, 0, content)


def _fixup_box_size(f, start_pos: int, box_type: bytes):
    """Fix up the size field of a box that was written with placeholder."""
    end_pos = f.tell()
    size = end_pos - start_pos
    f.seek(start_pos)
    f.write(struct.pack(">I", size))
    f.write(box_type)
    f.seek(end_pos)


def _write_mdat(f, data: bytes):
    """Write mdat box with decrypted data."""
    size = len(data) + 8
    f.write(struct.pack(">I", size))
    f.write(b"mdat")
    f.write(data)


def _extract_stsd_content(data: bytes) -> Optional[bytes]:
    """Extract full stsd box content from moov box (supports any codec)."""
    # Find stsd box in the data
    idx = data.find(b"stsd")
    if idx < 4:
        return None

    # Get stsd box size
    size = struct.unpack(">I", data[idx - 4 : idx])[0]
    if size < 16 or size > 10000:  # Reasonable stsd size range
        return None

    # Return stsd content (after box header = size + type)
    raw_content = data[idx + 4 : idx - 4 + size]

    # Clean the stsd content to remove encryption metadata
    return _clean_stsd_content(raw_content)


def _clean_stsd_content(stsd_content: bytes) -> bytes:
    """
    Clean stsd content by removing encryption metadata.

    This replaces the mp4decrypt cleanup step. For encrypted files, the stsd
    contains encrypted sample entries (enca, encv) with sinf boxes that describe
    the encryption scheme. We need to:
    1. Convert encrypted entries (enca -> original format from frma)
    2. Remove sinf boxes entirely
    """
    if len(stsd_content) < 8:
        return stsd_content

    # stsd content: version(1) + flags(3) + entry_count(4) + entries...
    version_flags = stsd_content[:4]
    entry_count = struct.unpack(">I", stsd_content[4:8])[0]

    # Parse and clean each sample entry
    cleaned_entries = []
    offset = 8

    for _ in range(entry_count):
        if offset + 8 > len(stsd_content):
            break

        entry_size = struct.unpack(">I", stsd_content[offset : offset + 4])[0]
        entry_type = stsd_content[offset + 4 : offset + 8]

        if entry_size < 8 or offset + entry_size > len(stsd_content):
            break

        entry_data = stsd_content[offset : offset + entry_size]

        # Check if this is an encrypted entry
        if entry_type in (b"enca", b"encv", b"encs", b"encm"):
            # Clean the encrypted entry
            cleaned_entry = _clean_encrypted_sample_entry(entry_data)
            cleaned_entries.append(cleaned_entry)
        else:
            # Keep as-is but still remove any sinf boxes that might be present
            cleaned_entry = _remove_sinf_from_entry(entry_data)
            cleaned_entries.append(cleaned_entry)

        offset += entry_size

    # Rebuild stsd content
    result = version_flags + struct.pack(">I", len(cleaned_entries))
    for entry in cleaned_entries:
        result += entry

    return result


def _clean_encrypted_sample_entry(entry_data: bytes) -> bytes:
    """
    Clean an encrypted sample entry (enca, encv, etc.).

    Structure of encrypted audio entry (enca):
    - Box header (8 bytes): size + 'enca'
    - Reserved (6 bytes)
    - Data reference index (2 bytes)
    - Audio specific data (varies by codec)
    - Child boxes including:
      - sinf (protection scheme info) - REMOVE THIS
      - Original codec box (alac, esds, etc.) - KEEP

    We need to:
    1. Find frma box inside sinf to get original format
    2. Replace 'enca' with original format
    3. Remove sinf box
    """
    if len(entry_data) < 36:  # Minimum audio sample entry size
        return entry_data

    entry_size = struct.unpack(">I", entry_data[:4])[0]
    entry_type = entry_data[4:8]

    # Find the original format from sinf/frma
    original_format = _find_original_format(entry_data)
    if not original_format:
        # If we can't find frma, try common mappings
        if entry_type == b"enca":
            original_format = b"mp4a"  # Default to AAC
        elif entry_type == b"encv":
            original_format = b"avc1"  # Default to H.264
        else:
            original_format = entry_type  # Keep as-is

    # Audio sample entry structure:
    # - size (4) + type (4) + reserved (6) + data_ref_index (2) + audio_data (20) = 36 bytes
    # - Then child boxes start at offset 36

    # Copy the fixed header part, replacing the type
    new_entry = entry_data[:4] + original_format + entry_data[8:36]

    # Process child boxes, removing sinf
    child_offset = 36
    while child_offset + 8 <= len(entry_data):
        child_size = struct.unpack(">I", entry_data[child_offset : child_offset + 4])[0]
        child_type = entry_data[child_offset + 4 : child_offset + 8]

        if child_size < 8 or child_offset + child_size > len(entry_data):
            break

        # Skip sinf box (encryption metadata)
        if child_type != b"sinf":
            new_entry += entry_data[child_offset : child_offset + child_size]

        child_offset += child_size

    # Update the entry size
    new_size = len(new_entry)
    new_entry = struct.pack(">I", new_size) + new_entry[4:]

    return new_entry


def _find_original_format(entry_data: bytes) -> Optional[bytes]:
    """
    Find the original format (frma) from sinf box in an encrypted entry.

    sinf box contains:
    - frma: original format (4 bytes codec type)
    - schm: scheme type
    - schi: scheme info (contains tenc for CENC)
    """
    # Look for sinf box
    sinf_idx = entry_data.find(b"sinf")
    if sinf_idx < 4:
        return None

    sinf_size = struct.unpack(">I", entry_data[sinf_idx - 4 : sinf_idx])[0]
    if sinf_size < 16 or sinf_idx + sinf_size > len(entry_data) + 4:
        return None

    sinf_data = entry_data[sinf_idx - 4 : sinf_idx - 4 + sinf_size]

    # Look for frma box inside sinf
    frma_idx = sinf_data.find(b"frma")
    if frma_idx < 4:
        return None

    frma_size = struct.unpack(">I", sinf_data[frma_idx - 4 : frma_idx])[0]
    if frma_size != 12:  # frma is always 12 bytes: size(4) + type(4) + format(4)
        return None

    # Extract the original format
    return sinf_data[frma_idx + 4 : frma_idx + 8]


def _remove_sinf_from_entry(entry_data: bytes) -> bytes:
    """
    Remove sinf box from a sample entry (if present).
    Used for non-encrypted entries that might still have protection info.
    """
    if len(entry_data) < 36:
        return entry_data

    # Check if sinf exists
    if b"sinf" not in entry_data:
        return entry_data

    # Rebuild entry without sinf
    new_entry = entry_data[:36]  # Keep header and audio data

    child_offset = 36
    while child_offset + 8 <= len(entry_data):
        child_size = struct.unpack(">I", entry_data[child_offset : child_offset + 4])[0]
        child_type = entry_data[child_offset + 4 : child_offset + 8]

        if child_size < 8 or child_offset + child_size > len(entry_data):
            break

        if child_type != b"sinf":
            new_entry += entry_data[child_offset : child_offset + child_size]

        child_offset += child_size

    # Update size
    new_size = len(new_entry)
    new_entry = struct.pack(">I", new_size) + new_entry[4:]

    return new_entry


def _extract_alac_config(data: bytes) -> Optional[bytes]:
    """Extract ALAC configuration from moov/stsd box (for backwards compatibility)."""
    # Simple search for 'alac' box in data
    idx = data.find(b"alac")
    if idx < 4:
        return None

    # Check if it's inside stsd (look for full structure)
    # The 'alac' cookie box follows the sample entry
    alac_idx = idx
    while alac_idx < len(data) - 100:
        if data[alac_idx : alac_idx + 4] == b"alac":
            size = struct.unpack(">I", data[alac_idx - 4 : alac_idx])[0]
            if 20 < size < 100:  # Reasonable ALAC config size
                return data[alac_idx + 4 : alac_idx - 4 + size]
        alac_idx += 1
        if alac_idx > idx + 200:
            break
    return None


def _extract_timescale(data: bytes) -> int:
    """Extract timescale from moov/mvhd or mdhd box."""
    # Look for mdhd box (media header has the audio timescale)
    idx = data.find(b"mdhd")
    if idx > 0 and idx + 24 < len(data):
        # mdhd: version(1) + flags(3) + creation(4) + modification(4) + timescale(4)
        return struct.unpack(">I", data[idx + 16 : idx + 20])[0]
    return 44100  # Default


def _find_child_box(
    container_data: bytes, target_type: bytes, skip_header: int = 8
) -> Optional[bytes]:
    """Find a direct child box in container data.

    Args:
        container_data: Raw bytes of the container box (including its own header).
        target_type: 4-byte box type to search for.
        skip_header: Bytes to skip at start (8 for plain box, 12 for FullBox).

    Returns:
        Full box bytes (size + type + content) or None.
    """
    offset = skip_header
    while offset + 8 <= len(container_data):
        size = struct.unpack(">I", container_data[offset : offset + 4])[0]
        box_type = container_data[offset + 4 : offset + 8]
        if size < 8 or offset + size > len(container_data):
            break
        if box_type == target_type:
            return container_data[offset : offset + size]
        offset += size
    return None


def _find_audio_trak(moov_data: bytes) -> Optional[bytes]:
    """Find the audio trak box in moov data.

    Iterates trak children and returns the first one whose hdlr has
    handler_type == 'soun'. Returns full trak box bytes or None.
    """
    offset = 8  # Skip moov header
    while offset + 8 <= len(moov_data):
        size = struct.unpack(">I", moov_data[offset : offset + 4])[0]
        box_type = moov_data[offset + 4 : offset + 8]
        if size < 8 or offset + size > len(moov_data):
            break
        if box_type == b"trak":
            trak_data = moov_data[offset : offset + size]
            hdlr_idx = trak_data.find(b"hdlr")
            if hdlr_idx > 0:
                # hdlr FullBox: version+flags(4) + pre_defined(4) + handler_type(4)
                handler_offset = hdlr_idx + 4 + 4 + 4
                if handler_offset + 4 <= len(trak_data):
                    if trak_data[handler_offset : handler_offset + 4] == b"soun":
                        return trak_data
        offset += size
    return None


def _patch_mvhd_duration(box_data: bytes, duration: int) -> bytes:
    """Return a copy of the mvhd box with its duration field patched."""
    data = bytearray(box_data)
    version = data[8]  # After size(4) + type(4)
    if version == 0:
        # v0: ver+flags(4) + creation(4) + modification(4) + timescale(4) + duration(4)
        struct.pack_into(">I", data, 24, duration)
    else:
        # v1: ver+flags(4) + creation(8) + modification(8) + timescale(4) + duration(8)
        struct.pack_into(">Q", data, 32, duration)
    return bytes(data)


def _patch_tkhd_duration(box_data: bytes, duration: int) -> bytes:
    """Return a copy of the tkhd box with duration patched and flags set to 7."""
    data = bytearray(box_data)
    version = data[8]
    # Set flags = 7 (enabled | in_movie | in_preview)
    data[9:12] = struct.pack(">I", 7)[1:]  # 3-byte flags
    if version == 0:
        # v0: ver+flags(4) + creation(4) + modification(4) + track_id(4) + reserved(4) + duration(4)
        struct.pack_into(">I", data, 28, duration)
    else:
        # v1: ver+flags(4) + creation(8) + modification(8) + track_id(4) + reserved(4) + duration(8)
        struct.pack_into(">Q", data, 36, duration)
    return bytes(data)


def _patch_mdhd_duration(box_data: bytes, duration: int) -> bytes:
    """Return a copy of the mdhd box with its duration field patched.

    Preserves the original language code and all other fields.
    """
    data = bytearray(box_data)
    version = data[8]
    if version == 0:
        # Same layout as mvhd v0
        struct.pack_into(">I", data, 24, duration)
    else:
        struct.pack_into(">Q", data, 32, duration)
    return bytes(data)


def _write_udta(f):
    """Write udta > meta > hdlr(mdir) + ilst metadata container.

    This matches the Go amdecrypt output which creates an empty metadata
    container so tools can find and write metadata atoms.
    """
    udta_start = f.tell()
    f.write(b"\x00" * 8)  # udta placeholder

    meta_start = f.tell()
    f.write(b"\x00" * 8)  # meta placeholder
    # meta is a FullBox: version(1) + flags(3)
    f.write(struct.pack(">I", 0))  # version + flags

    # hdlr for metadata (handler_type = 'mdir', reserved = 'appl' + zeros)
    hdlr_content = struct.pack(">I", 0)  # pre_defined
    hdlr_content += b"mdir"  # handler_type
    hdlr_content += struct.pack(">III", 0x6170706C, 0, 0)  # reserved ('appl', 0, 0)
    hdlr_content += b"\x00"  # empty name (null terminator)
    _write_fullbox(f, b"hdlr", 0, 0, hdlr_content)

    # ilst (empty)
    _write_box(f, b"ilst", b"")

    _fixup_box_size(f, meta_start, b"meta")
    _fixup_box_size(f, udta_start, b"udta")


def _extract_encryption_info(moov_data: bytes) -> Optional[EncryptionInfo]:
    """Extract encryption scheme info from the audio track's sinf box.

    Walks moov → trak (audio) → mdia → minf → stbl → stsd → sample_entry → sinf,
    then reads sinf/schm for scheme_type and sinf/schi/tenc for IV size, constant IV,
    and default KID.

    Returns EncryptionInfo or None if no sinf is found.
    """
    trak_data = _find_audio_trak(moov_data)
    if trak_data is None:
        return None

    # Navigate trak → mdia → minf → stbl → stsd
    mdia = _find_child_box(trak_data, b"mdia")
    if mdia is None:
        return None
    minf = _find_child_box(mdia, b"minf")
    if minf is None:
        return None
    stbl = _find_child_box(minf, b"stbl")
    if stbl is None:
        return None
    stsd = _find_child_box(stbl, b"stsd")
    if stsd is None:
        return None

    # stsd is a FullBox: 4 (size) + 4 (type) + 4 (version+flags) + 4 (entry_count)
    # Then the first sample entry immediately follows
    if len(stsd) < 16:
        return None
    entry_offset = 16  # past header+version+flags+entry_count
    if entry_offset + 8 > len(stsd):
        return None
    entry_size = struct.unpack(">I", stsd[entry_offset : entry_offset + 4])[0]
    entry_data = stsd[entry_offset : entry_offset + entry_size]

    # Find sinf inside this sample entry
    # Audio sample entries have a 36-byte fixed header:
    #   size(4) + type(4) + reserved(6) + data_ref_index(2) + audio_data(20)
    # Child boxes (including sinf) start at offset 36
    sinf = _find_child_box(entry_data, b"sinf", skip_header=36)
    if sinf is None:
        return None

    info = EncryptionInfo()

    # Parse schm (Scheme Type Box) inside sinf
    schm = _find_child_box(sinf, b"schm")
    if schm and len(schm) >= 20:
        # schm: 4(size) + 4(type) + 4(ver+flags) + 4(scheme_type) + 4(scheme_version)
        info.scheme_type = schm[12:16].decode("ascii", errors="replace")
        logger.debug(f"Encryption scheme: {info.scheme_type}")

    # Parse tenc (Track Encryption Box) inside sinf/schi
    schi = _find_child_box(sinf, b"schi")
    if schi:
        tenc = _find_child_box(schi, b"tenc")
        if tenc and len(tenc) >= 32:
            # tenc FullBox layout (offsets include 8-byte box header):
            #   [0:4]  size
            #   [4:8]  type "tenc"
            #   [8]    version
            #   [9:12] flags (3 bytes)
            #   [12]   reserved
            #   [13]   reserved (v0) / crypt_byte_block|skip_byte_block (v1)
            #   [14]   default_isProtected
            #   [15]   default_Per_Sample_IV_Size
            #   [16:32] default_KID (16 bytes)
            #   if per_sample_iv_size==0:
            #     [32]   default_constant_IV_size
            #     [33..] default_constant_IV
            tenc_version = tenc[8]
            per_sample_iv_size = tenc[15]
            kid = tenc[16:32]

            info.per_sample_iv_size = per_sample_iv_size
            info.kid = kid
            logger.debug(
                f"tenc: per_sample_iv_size={per_sample_iv_size}, " f"kid={kid.hex()}"
            )

            # If per_sample_iv_size is 0, a constant IV follows the KID
            if per_sample_iv_size == 0 and len(tenc) > 32:
                constant_iv_size = tenc[32]
                if len(tenc) >= 33 + constant_iv_size:
                    info.constant_iv = tenc[33 : 33 + constant_iv_size]
                    logger.debug(f"Constant IV: {info.constant_iv.hex()}")

    return info


def _extract_encryption_info_per_stsd(moov_data: bytes) -> Optional[dict]:
    """Extract encryption scheme info for each stsd entry (sample description).

    Returns a dict mapping desc_index (0-based) → EncryptionInfo, or None if no
    encryption found. This handles cases where different sample descriptions have
    different encryption parameters (e.g., different IVs or key schemes).
    """
    trak_data = _find_audio_trak(moov_data)
    if trak_data is None:
        return None

    # Navigate trak → mdia → minf → stbl → stsd
    mdia = _find_child_box(trak_data, b"mdia")
    if mdia is None:
        return None
    minf = _find_child_box(mdia, b"minf")
    if minf is None:
        return None
    stbl = _find_child_box(minf, b"stbl")
    if stbl is None:
        return None
    stsd = _find_child_box(stbl, b"stsd")
    if stsd is None:
        return None

    if len(stsd) < 16:
        return None

    entry_count = struct.unpack(">I", stsd[12:16])[0]
    if entry_count == 0:
        return None

    encryption_info_per_desc = {}
    entry_offset = 16  # past header+version+flags+entry_count

    for desc_idx in range(entry_count):
        if entry_offset + 8 > len(stsd):
            break

        entry_size = struct.unpack(">I", stsd[entry_offset : entry_offset + 4])[0]
        if entry_size < 8 or entry_offset + entry_size > len(stsd):
            break

        entry_data = stsd[entry_offset : entry_offset + entry_size]

        # Find sinf inside this sample entry
        # Audio sample entries have 36-byte fixed header before child boxes
        sinf = _find_child_box(entry_data, b"sinf", skip_header=36)
        if sinf is not None:
            # Extract encryption info for this stsd entry
            info = EncryptionInfo()

            # Parse schm
            schm = _find_child_box(sinf, b"schm")
            if schm and len(schm) >= 20:
                info.scheme_type = schm[12:16].decode("ascii", errors="replace")
                logger.debug(
                    f"Encryption scheme for desc_index {desc_idx}: {info.scheme_type}"
                )

            # Parse tenc
            schi = _find_child_box(sinf, b"schi")
            if schi:
                tenc = _find_child_box(schi, b"tenc")
                if tenc and len(tenc) >= 32:
                    per_sample_iv_size = tenc[15]
                    kid = tenc[16:32]

                    info.per_sample_iv_size = per_sample_iv_size
                    info.kid = kid
                    logger.debug(
                        f"tenc (desc {desc_idx}): per_sample_iv_size={per_sample_iv_size}"
                    )

                    # If per_sample_iv_size is 0, extract constant IV
                    if per_sample_iv_size == 0 and len(tenc) > 32:
                        constant_iv_size = tenc[32]
                        if len(tenc) >= 33 + constant_iv_size:
                            info.constant_iv = tenc[33 : 33 + constant_iv_size]
                            logger.debug(
                                f"Constant IV (desc {desc_idx}): {info.constant_iv.hex()}"
                            )

            encryption_info_per_desc[desc_idx] = info

        entry_offset += entry_size

    return encryption_info_per_desc if encryption_info_per_desc else None


def _extract_audio_track_id(moov_data: bytes) -> int:
    """Extract the track ID of the audio track from the moov box.

    Parses trak boxes in moov to find one with handler_type 'soun' (sound),
    then returns its track_id from tkhd. Defaults to 1 if not found.
    """
    offset = 8  # Skip moov box header
    while offset < len(moov_data) - 8:
        size = struct.unpack(">I", moov_data[offset : offset + 4])[0]
        box_type = moov_data[offset + 4 : offset + 8]

        if size < 8 or offset + size > len(moov_data):
            break

        if box_type == b"trak":
            trak_data = moov_data[offset : offset + size]

            # Check handler type in hdlr box
            hdlr_idx = trak_data.find(b"hdlr")
            if hdlr_idx > 0:
                # hdlr FullBox: after 'hdlr' type comes version+flags(4) + pre_defined(4) + handler_type(4)
                handler_offset = hdlr_idx + 4 + 4 + 4
                if handler_offset + 4 <= len(trak_data):
                    handler_type = trak_data[handler_offset : handler_offset + 4]
                    if handler_type == b"soun":
                        # Found audio track, extract track_id from tkhd
                        tkhd_idx = trak_data.find(b"tkhd")
                        if tkhd_idx > 0:
                            version = trak_data[tkhd_idx + 4]
                            if version == 0:
                                # v0: ver+flags(4) + creation(4) + modification(4) + track_id(4)
                                tid_offset = tkhd_idx + 4 + 4 + 4 + 4
                            else:
                                # v1: ver+flags(4) + creation(8) + modification(8) + track_id(4)
                                tid_offset = tkhd_idx + 4 + 4 + 8 + 8
                            if tid_offset + 4 <= len(trak_data):
                                return struct.unpack(
                                    ">I", trak_data[tid_offset : tid_offset + 4]
                                )[0]

        offset += size

    return 1  # Default to track 1


async def decrypt_file(
    wrapper_ip: str,
    track_id: str,
    fairplay_key: str,
    input_path: str,
    output_path: str,
    progress_callback=None,
) -> None:
    """
    Main decryption function - decrypt an encrypted MP4 file via the wrapper.

    This is the Python equivalent of the amdecrypt tool:
    1. Extract samples from encrypted MP4
    2. Send samples to wrapper for FairPlay decryption
    3. Reassemble decrypted MP4 with clean metadata

    Args:
        wrapper_ip: Wrapper decrypt port address (e.g., "127.0.0.1:10020")
        track_id: Apple Music track ID
        fairplay_key: FairPlay key URI (skd://...)
        input_path: Path to encrypted MP4 file
        output_path: Path for decrypted output file
        progress_callback: Optional callback(current, total, bytes, speed) for decryption progress
    """
    logger.debug(f"Decrypting {input_path} -> {output_path}")

    # Extract samples (run in thread to not block)
    song_info = await asyncio.to_thread(extract_song, input_path)

    # Decrypt samples via wrapper
    decrypted_data = await decrypt_samples(
        wrapper_ip,
        track_id,
        fairplay_key,
        song_info.samples,
        progress_callback,
    )

    # Write output file (preserve original structure, replace mdat content)
    # Encryption metadata is automatically cleaned during stsd extraction
    await asyncio.to_thread(
        write_decrypted_m4a,
        output_path,
        song_info,
        decrypted_data,
        input_path,  # Pass original path for codec info extraction
    )


def decrypt_samples_hex(
    samples: List[SampleInfo],
    keys: dict,
    encryption_info: EncryptionInfo,
    encryption_info_per_desc: Optional[dict] = None,
) -> bytes:
    """Decrypt samples using hex AES keys (no wrapper needed).

    Supports both CENC (AES-128-CTR) and CBCS (AES-128-CBC) schemes.

    Args:
        samples: List of SampleInfo with data, desc_index, iv, subsamples.
        keys: Mapping of desc_index (int) → AES key (16 bytes, raw).
        encryption_info: EncryptionInfo with scheme_type, constant_iv, etc.
        encryption_info_per_desc: Optional dict mapping desc_index → EncryptionInfo
                                   (used when different stsd entries have different params).

    Returns:
        Concatenated decrypted sample data.
    """
    is_cenc = encryption_info.scheme_type == "cenc"
    decrypted = bytearray()

    for sample in samples:
        key = keys.get(sample.desc_index)
        if key is None:
            # No key for this desc_index — keep data as-is (shouldn't happen)
            decrypted.extend(sample.data)
            continue

        # Get encryption info for this sample's desc_index (if per-description info exists)
        if encryption_info_per_desc and sample.desc_index in encryption_info_per_desc:
            enc_info = encryption_info_per_desc[sample.desc_index]
        else:
            enc_info = encryption_info

        if is_cenc:
            # AES-128-CTR: per-sample IV from senc, zero-padded to 16 bytes
            iv = sample.iv
            if len(iv) < 16:
                iv = iv + b"\x00" * (16 - len(iv))
            cipher = AES.new(key, AES.MODE_CTR, nonce=b"", initial_value=iv)

            if sample.subsamples:
                plaintext = bytearray()
                offset = 0
                for clear_bytes, encrypted_bytes in sample.subsamples:
                    plaintext.extend(sample.data[offset : offset + clear_bytes])
                    offset += clear_bytes
                    plaintext.extend(
                        cipher.decrypt(sample.data[offset : offset + encrypted_bytes])
                    )
                    offset += encrypted_bytes
                plaintext.extend(sample.data[offset:])
                decrypted.extend(plaintext)
            else:
                decrypted.extend(cipher.decrypt(sample.data))

        else:
            # CBCS (AES-128-CBC): constant IV or per-sample IV
            iv = sample.iv if sample.iv else enc_info.constant_iv
            if len(iv) < 16:
                iv = iv + b"\x00" * (16 - len(iv))

            if sample.subsamples:
                # For CBCS subsamples: concatenate all encrypted regions into one,
                # decrypt as one CBC stream (to maintain cipher state), then split back.
                # This avoids losing bytes if encrypted_bytes values aren't 16-byte aligned.

                # Collect all encrypted byte ranges and encrypt content
                encrypted_concat = bytearray()
                subsample_sizes = (
                    []
                )  # Track size of each encrypted region for reassembly
                offset = 0
                for clear_bytes, encrypted_bytes in sample.subsamples:
                    offset += clear_bytes
                    if encrypted_bytes > 0:
                        encrypted_concat.extend(
                            sample.data[offset : offset + encrypted_bytes]
                        )
                        subsample_sizes.append(encrypted_bytes)
                    offset += encrypted_bytes

                # Decrypt concatenated regions as one CBC stream
                total_enc_len = len(encrypted_concat)
                decrypted_concat = bytearray()
                if total_enc_len > 0:
                    cbc_len = total_enc_len & ~0xF
                    if cbc_len > 0:
                        cipher = AES.new(key, AES.MODE_CBC, iv=iv)
                        decrypted_concat.extend(
                            cipher.decrypt(bytes(encrypted_concat[:cbc_len]))
                        )
                    # Any trailing unaligned bytes (shouldn't happen if file is well-formed)
                    if cbc_len < total_enc_len:
                        decrypted_concat.extend(encrypted_concat[cbc_len:])

                # Reassemble with original clear/encrypted pattern
                plaintext = bytearray()
                dec_offset = 0
                offset = 0
                for clear_bytes, encrypted_bytes in sample.subsamples:
                    plaintext.extend(sample.data[offset : offset + clear_bytes])
                    offset += clear_bytes
                    if encrypted_bytes > 0:
                        plaintext.extend(
                            decrypted_concat[dec_offset : dec_offset + encrypted_bytes]
                        )
                        dec_offset += encrypted_bytes
                    offset += encrypted_bytes
                plaintext.extend(sample.data[offset:])
                decrypted.extend(plaintext)
            else:
                # Full subsample: for well-formed files, the entire sample should be
                # a multiple of 16 bytes. Only truncate if misaligned (unexpected).
                sample_len = len(sample.data)
                if sample_len % 16 == 0:
                    # Data is properly 16-byte aligned, decrypt as-is
                    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
                    decrypted.extend(cipher.decrypt(sample.data))
                else:
                    # Data is not aligned (unexpected case) - truncate carefully
                    truncated_len = sample_len & ~0xF
                    if truncated_len > 0:
                        cipher = AES.new(key, AES.MODE_CBC, iv=iv)
                        decrypted.extend(cipher.decrypt(sample.data[:truncated_len]))
                        # Keep unaligned tail bytes as clear (unencrypted)
                        decrypted.extend(sample.data[truncated_len:])
                    else:
                        # Less than 16 bytes - cannot decrypt with CBC, keep as-is
                        decrypted.extend(sample.data)

    logger.debug(
        f"Decrypted {len(samples)} samples ({len(decrypted)} bytes) with hex keys"
    )
    return bytes(decrypted)


async def decrypt_file_hex(
    input_path: str,
    output_path: str,
    decryption_key: str,
    legacy: bool = False,
) -> None:
    """Decrypt an encrypted MP4 file using a hex AES key (no wrapper/mp4decrypt).

    This replaces the mp4decrypt + remux pipeline with pure-Python decryption:
    1. Extract samples and encryption info from MP4
    2. Decrypt samples using AES (CTR for cenc / CBC for cbcs)
    3. Write clean decrypted M4A output

    Args:
        input_path: Path to encrypted MP4 file.
        output_path: Path for decrypted output file.
        decryption_key: Hex-encoded 128-bit AES key (32 hex chars).
        legacy: If True, treat as legacy AAC (cenc, single key).
    """
    logger.debug(f"Hex-key decrypt: {input_path} -> {output_path}")

    # Extract samples (run in thread to not block)
    song_info = await asyncio.to_thread(extract_song, input_path)

    # Build key mapping: desc_index → raw AES key bytes
    track_key = bytes.fromhex(decryption_key)

    if legacy:
        # Legacy AAC (cenc): single key for all samples (all desc_index 0)
        keys = {0: track_key}
    else:
        # Non-legacy (cbcs): two sample descriptions
        #   desc_index 0 → DEFAULT_SONG_DECRYPTION_KEY (prefetch samples)
        #   desc_index 1 → track key (from Widevine CDM)
        keys = {0: DEFAULT_SONG_DECRYPTION_KEY, 1: track_key}

    # Use encryption info from the file (fall back to sensible defaults)
    enc_info = song_info.encryption_info or EncryptionInfo(
        scheme_type="cenc" if legacy else "cbcs"
    )

    # Try to extract per-description encryption info (for non-legacy files)
    # This handles cases where desc_index 0 and 1 have different encryption parameters
    enc_info_per_desc = None
    if song_info.moov_data and not legacy:
        enc_info_per_desc = await asyncio.to_thread(
            _extract_encryption_info_per_stsd, song_info.moov_data
        )
        if enc_info_per_desc:
            logger.debug(
                f"Found per-description encryption info: {list(enc_info_per_desc.keys())}"
            )

    # Decrypt
    decrypted_data = decrypt_samples_hex(
        song_info.samples, keys, enc_info, enc_info_per_desc
    )

    # Write output (preserves original metadata boxes)
    await asyncio.to_thread(
        write_decrypted_m4a,
        output_path,
        song_info,
        decrypted_data,
        input_path,
    )



import asyncio
import json
import string
import subprocess
import typing

import httpx


def raise_for_status(httpx_response: httpx.Response, valid_responses: set[int] = {200}):
    if httpx_response.status_code not in valid_responses:
        raise httpx._exceptions.HTTPError(
            f"HTTP error {httpx_response.status_code}: {httpx_response.text}"
        )


def safe_json(httpx_response: httpx.Response) -> dict | None:
    try:
        return httpx_response.json()
    except (json.JSONDecodeError, UnicodeDecodeError):
        return None


async def get_response(
    url: str,
    valid_responses: set[int] = {200},
) -> httpx.Response:
    async with httpx.AsyncClient(timeout=60.0) as client:
        response = await client.get(url)
        raise_for_status(response, valid_responses)
        return response


async def async_subprocess(*args: str, silent: bool = False) -> None:
    if silent:
        additional_args = {
            "stdout": subprocess.DEVNULL,
            "stderr": subprocess.DEVNULL,
        }
    else:
        additional_args = {}

    proc = await asyncio.create_subprocess_exec(
        *args,
        **additional_args,
    )
    await proc.communicate()

    if proc.returncode != 0:
        raise Exception(f'"{args[0]}" exited with code {proc.returncode}')


async def safe_gather(
    *tasks: typing.Awaitable[typing.Any],
    limit: int = 10,
) -> list[typing.Any]:
    semaphore = asyncio.Semaphore(limit)

    async def bounded_task(task: typing.Awaitable[typing.Any]) -> typing.Any:
        async with semaphore:
            return await task

    return await asyncio.gather(
        *(bounded_task(task) for task in tasks),
        return_exceptions=True,
    )


async def sequential_gather(
    *tasks: typing.Awaitable[typing.Any],
    interval: float = 0.5,
) -> list[typing.Any]:
    results = []
    for i, task in enumerate(tasks):
        try:
            result = await task
            results.append(result)
        except Exception as e:
            results.append(e)
        if interval > 0 and i < len(tasks) - 1:
            await asyncio.sleep(interval)
    return results


class CustomStringFormatter(string.Formatter):
    def format_field(self, value: typing.Any, format_spec: str) -> str:
        if isinstance(value, tuple) and len(value) == 2:
            actual_value, fallback_value = value
            if actual_value is None:
                return fallback_value

            try:
                return super().format_field(actual_value, format_spec)
            except Exception:
                return fallback_value

        return super().format_field(value, format_spec)


class GamdlError(Exception):
    pass



from enum import Enum

from .constants import (
    FOURCC_MAP,
    LEGACY_SONG_CODECS,
    MEDIA_RATING_STR_MAP,
    MEDIA_TYPE_STR_MAP,
)


class SyncedLyricsFormat(Enum):
    LRC = "lrc"
    SRT = "srt"
    TTML = "ttml"


class MediaType(Enum):
    SONG = 1
    MUSIC_VIDEO = 6

    def __str__(self) -> str:
        return MEDIA_TYPE_STR_MAP[self.value]

    def __int__(self) -> int:
        return self.value


class MediaRating(Enum):
    NONE = 0
    EXPLICIT = 1
    CLEAN = 2

    def __str__(self) -> str:
        return MEDIA_RATING_STR_MAP[self.value]

    def __int__(self) -> int:
        return self.value


class MediaFileFormat(Enum):
    MP4 = "mp4"
    M4V = "m4v"
    M4A = "m4a"


class SongCodec(Enum):
    AAC_LEGACY = "aac-legacy"
    AAC_HE_LEGACY = "aac-he-legacy"
    AAC = "aac"
    AAC_HE = "aac-he"
    AAC_BINAURAL = "aac-binaural"
    AAC_DOWNMIX = "aac-downmix"
    AAC_HE_BINAURAL = "aac-he-binaural"
    AAC_HE_DOWNMIX = "aac-he-downmix"
    ATMOS = "atmos"
    AC3 = "ac3"
    ALAC = "alac"
    ASK = "ask"

    def is_legacy(self) -> bool:
        return self.value in LEGACY_SONG_CODECS


class MusicVideoCodec(Enum):
    H264 = "h264"
    H265 = "h265"
    ASK = "ask"

    def fourcc(self) -> str:
        return FOURCC_MAP[self.value]


class MusicVideoResolution(Enum):
    R240P = "240p"
    R360P = "360p"
    R480P = "480p"
    R540P = "540p"
    R720P = "720p"
    R1080P = "1080p"
    R1440P = "1440p"
    R2160P = "2160p"

    def __int__(self) -> int:
        return int(self.value[:-1])


class UploadedVideoQuality(Enum):
    BEST = "best"
    ASK = "ask"


class CoverFormat(Enum):
    JPG = "jpg"
    PNG = "png"
    RAW = "raw"
